{-# LANGUAGE LambdaCase, TypeFamilies, CPP #-}

module Expr where

-- November 30, 2020

import Data.Maybe
import Control.Exception
import Control.Applicative (Applicative(pure,(<*>)),Alternative(empty,(<|>)))
import Control.Monad
import Painter (readFileContinue,fold2,mkInt,Rules,reduce,update,updList,
	        MaybeT(MaybeT,runMT),MonadTrans(type M,lift,liftT),Tree(V,F),
	        root,mkTree,insert)
import Coalg (TreeSig(TS,var,fun),foldTree,State(State,runS),single,parse,
	      StateT(StateT,runST),Compiler,sat,char,token,tchar,some,many,nat)

#if __GLASGOW_HASKELL__ >= 806
import Control.Monad.Fail hiding (fail)		
import qualified Control.Monad.Fail (fail)	
instance MonadFail m => MonadFail (ListT m) where fail = fail
#endif

-- monad transformers

newtype ListT m a = ListT {runLT :: m [a]}

instance Monad m => Monad (ListT m) where
         return = ListT . return . single
 	 m >>= f = ListT $ do mas <- runLT m
 	 		      mass <- mapM (runLT . f) mas
 			      return $ concat mass
         fail _ = mzero		   
			       		        
instance Monad m => MonadPlus (ListT m) where
 	 mzero = ListT $ return []
         m `mplus` m' = ListT $ do mas <- runLT m; mas' <- runLT m'
 			           return $ mas++mas'

instance Monad m => Functor (ListT m) where fmap = liftM 
 	 
instance Monad m => Applicative (ListT m) where pure  = return
         					(<*>) = liftM2 id

instance Monad m => Alternative (ListT m) where empty = mzero
 	 					(<|>) = mplus
		 
instance MonadTrans ListT where 
		    type M ListT = [ ] 
		    lift  = ListT . liftM single -- :: m a -> ListT m a
		    liftT = ListT . return	 -- :: [a] -> ListT m a

-- basic compilers

nchar :: String -> Compiler Char
nchar chrs = sat (`notElem` chrs)

letter :: Compiler Char
letter = msum $ map char $ ['a'..'z']++['A'..'Z']

string :: String -> Compiler String
string = mapM char

bool :: Compiler Bool
bool = msum [string "True"  >> return True,
             string "False" >> return False]

int :: Compiler Int
int = msum [nat, do char '-'; n <- nat; return $ -n]

identifier :: Compiler String
identifier = liftM2 (:) letter $ many $ nchar "(){};=!>+-*/^ \n\t"

tstring     = token . string
tbool       = token bool
tint        = token int
tidentifier = token identifier
 	 					 
-- arithmetic expressions

data Exp x = Con Int | Var x | Sum [Exp x] | Prod [Exp x] | Exp x :- Exp x | 
	     Exp x :/ Exp x | Int :* Exp x | Exp x :^ Int deriving Eq

zero,one :: Exp x
zero = Con 0
one  = Con 1

instance Functor Exp where 
         fmap f = \case Con i   -> Con i
         		Var x   -> Var $ f x
         		Sum es  -> Sum $ map (fmap f) es
         		Prod es -> Prod $ map (fmap f) es
         		e :- e' -> fmap f e :- fmap f e'
         		e :/ e' -> fmap f e :/ fmap f e'
         		i :* e  -> i :* fmap f e
         		e :^ i  -> fmap f e :^ i
	    
instance Monad Exp where 					-- Exp monad
         return = Var
         e >>= f = case e of Con i   -> Con i
         		     Var x   -> f x
                             Sum es  -> Sum $ map (>>= f) es
                             Prod es -> Prod $ map (>>= f) es
                             e :- e' -> (e >>= f) :- (e' >>= f)
                             e :/ e' -> (e >>= f) :/ (e' >>= f)
                             i :* e  -> i :* (e >>= f)
                             e :^ i  -> (e >>= f) :^ i

instance Applicative Exp where pure  = return
        		       (<*>) = liftM2 id
	    
data BExp x = True_ | False_ | BVar x | Or [BExp x] | And [BExp x] |
              Not (BExp x) | Exp x :< Exp x | Exp x := Exp x | Exp x :<= Exp x
              deriving (Eq,Show)
	
type Store x = x -> Int

-- def/use example

data DefUse x = Def x (Exp x) | Use x
	 
trace :: Eq x => [DefUse x] -> Store x -> ([(x,Int)],Store x)
trace (Def x e:s) store = trace s $ updStore x e store
trace (Use x:s) store   = ((x,store x):s',store')
			  where (s',store') = trace s store
trace _ store	        = ([],store)

updStore :: Eq x => x -> Exp x -> Store x -> Store x
updStore x e store = update store x $ eval e store

put :: state -> State state ()
put state = State $ const ((),state)

get :: State state state
get = State $ \state -> (state,state)
	       
traceS :: Eq x => [DefUse x] -> State (Store x) [(x,Int)]  
traceS (Def x e:s) = do store <- get; put $ updStore x e store; traceS s
traceS (Use x:s)   = do store <- get; s <- traceS s; return $ (x,store x):s
traceS _           = return []
	       
traceIO :: (Eq x,Show x) => [DefUse x] -> StateT (Store x) IO ()  
traceIO (Def x e:s) = do store <- liftT get; liftT $ put $ updStore x e store
			 traceIO s
traceIO (Use x:s)   = do store <- liftT get; lift $ putStrLn $ show (x,store x)
		         traceIO s
traceIO _           = return ()
     
data V = X | Y | Z deriving (Eq,Show)

dflist :: [DefUse V]
dflist = [Def X $ Con 1,Use X,Def Y $ Con 2,Use Y,Def X $ Sum [Var X,Var Y],
	  Use X,Use Y]
		  
tra,traS :: [(V,Int)]
tra  = fst $ trace dflist $ const 0	 	-- > [(X,1),(Y,2),(X,3),(Y,2)]    
traS = fst $ runS (traceS dflist) $ const 0 	-- > [(X,1),(Y,2),(X,3),(Y,2)] 

traIO :: IO ((),Store V)
traIO = runST (traceIO dflist) $ const 0 	-- > (X,1)
						--   (Y,2)
						--   (X,3)
						--   (Y,2)
-- evaluation of arithmetic expressions

eval :: Exp x -> Store x -> Int	    
eval e st = case e of Con i   -> i
	              Var x   -> st x
	              Sum es  -> sum $ map f es
	              Prod es -> product $ map f es
	              e :- e' -> f e - f e'
	              e :/ e' -> f e `div` f e'
	              i :* e  -> i * f e
	              e :^ i  -> f e ^ i    
	    where f = flip eval st	  

evalR :: Exp x -> Store x -> Int 	            -- eval using the fact that 
evalR = \case Con i   -> return i	            -- (->) (Store x) is a monad
	      Var x   -> ($x)
	      Sum es  -> fmap sum $ mapM evalR es
	      Prod es -> fmap product $ mapM evalR es
	      e :- e' -> liftM2 (-) (evalR e) $ evalR e'
	      e :/ e' -> liftM2 div (evalR e) $ evalR e'
	      i :* e  -> liftM2 (-) (const i) $ evalR e
	      e :^ i  -> liftM2 (^) (evalR e) $ const i
	                             
writeVal :: Show x => Exp x -> Store x -> (String,Int)   -- employing  
writeVal e st = case e of			         -- the writer functor 
 		     Con i   -> ("",i)			 -- (,) String
		     Var x   -> out $ st x
		     Sum es  -> do is <- mapM f es; out $ sum is
		     Prod es -> do is <- mapM f es; out $ product is
		     e :- e' -> do i <- f e; k <- f e'; out $ i-k
		     e :/ e' -> do i <- f e; k <- f e'; out $ i`div`k
		     i :* e  -> do k <- f e; out $ i*k
		     e :^ i  -> do k <- f e; out $ k^i
                where out i = ('\n':filter (/='\"') (show e)++" = "++show i,i)
	              f = flip writeVal st
{-
writeVal e st = case e of			          
 		     Con i   -> ("",i)			 
		     Var x   -> out $ st x
		     Sum es  -> fmap sum (mapM f es) >>= out
		     Prod es -> fmap product (mapM f es) >>= out
		     e :- e' -> liftM2 (-) (f e) (f e') >>= out
		     e :/ e' -> liftM2 div (f e) (f e') >>= out
		     i :* e  -> fmap (i*) (f e) >>= out
		     e :^ i  -> fmap (^i) (f e) >>= out
                where out i = ('\n':filter (/='\"') (show e)++" = "++show i,i)
	              f = flip writeVal st
-} 
evalM :: Exp x -> Store x -> Maybe Int	  
evalM e st = case e of Con i   -> Just i
	               Var x   -> Just $ st x
	               Sum es  -> fmap sum $ mapM f es
	               Prod es -> fmap product $ mapM f es
	               e :- e' -> liftM2 (-) (f e) $ f e'
	               e :/ e' -> do k <- m; guard $ k /= 0; liftM2 div (f e) m
	         	          where m = f e'
                       i :* e  -> fmap (i*) $ f e
	               e :^ i  -> fmap (^i) $ f e
	     where f = flip evalM st	  

evalIO :: Exp String -> String -> MaybeT IO Int	    
evalIO e file = case e of Con i   -> return i
	                  Var x   -> do store <- lift $ readFileContinue [] file 
	            			              $ return . read
	            		        let val = lookup x store
	                                lift $ putStrLn $ x ++
	                                   case val of Just i -> " = " ++ show i
	                            		       _ -> " is undefined"
	                                liftT val
	                  Sum es  -> fmap sum $ mapM f es
	                  Prod es -> fmap product $ mapM f es
	                  e :- e' -> liftM2 (-) (f e) $ f e'
	                  e :/ e' -> do let m = f e'
	                                k <- m 
	       		                if k /= 0 then liftM2 div (f e) m
	       		                else do lift $ putStrLn 
	       		                                      "e/0 is undefined"
	       		                        mzero
	                  i :* e  -> fmap (i*) $ f e
	                  e :^ i  -> fmap (^i) $ f e
	        where f = flip evalIO file

-- arithmetic signatures

data Arith x val = Arith {con       :: Int -> val,
			  var_      :: x -> val,
			  sum_,prod :: [val] -> val,
			  sub,div_  :: val -> val -> val,
			  scal      :: Int -> val -> val,
			  expo      :: val -> Int -> val}

data ArithP x val = ArithP {conP       :: Int -> val,		
			    varP       :: x -> val,
			    sumP,prodP :: [Exp x] -> [val] -> val,
			    subP,divP  :: Exp x -> val -> Exp x -> val -> val,
			    scalP      :: Int -> Exp x -> val -> val,
			    expoP      :: Exp x -> val -> Int -> val}

-- translator from ArithP to Arith

mkArith :: ArithP x val -> Arith x (Exp x,val)
mkArith alg = Arith {con  = \i -> (Con i,conP alg i),
		     var_ = \x -> (Var x,varP alg x),	
		     sum_ = list Sum sumP,
		     prod = list Prod prodP,
		     sub  = bin (:-) subP,
		     div_ = bin (:/) divP,
		     scal = \i (e,a) -> (i:*e,scalP alg i e a),
		     expo = \(e,a) i -> (e:^i,expoP alg e a i)}
              where list c op eas = (c es,op alg es as) 
              			    where (es,as) = unzip eas
              	    bin c op (e,a) (e',b) = (c e e',op alg e a e' b)

-- catamorphic fold into an Arith-algebra

foldArith :: Arith x val -> Exp x -> val
foldArith alg = \case Con i   -> con alg i
		      Var x   -> var_ alg x
		      Sum es  -> sum_ alg $ map f es
		      Prod es -> prod alg $ map f es
		      e :- e' -> sub alg (f e) $ f e'
		      e :/ e' -> div_ alg (f e) $ f e'
		      i :* e  -> scal alg i $ f e
		      e :^ i  -> expo alg (f e) i
	        where f = foldArith alg

-- paramorphic fold into an ArithP-algebra

foldArithP :: ArithP x val -> Exp x -> val
foldArithP alg = \case Con i   -> conP alg i
		       Var x   -> varP alg x
		       Sum es  -> sumP alg es $ map f es
		       Prod es -> prodP alg es $ map f es
		       e :- e' -> subP alg e (f e) e' $ f e'
		       e :/ e' -> divP alg e (f e) e' $ f e'
		       i :* e  -> scalP alg i e $ f e
		       e :^ i  -> expoP alg e (f e) i
	         where f = foldArithP alg

-- Arith(P)-algebras

arithT :: Arith x (Exp x)		    -- term algebra
arithT = Arith Con Var Sum Prod (:-) (:/) (:*) (:^) 

evalAlg :: Arith x (Store x -> Int)    	    -- foldArith evalAlg = eval
evalAlg = Arith {con  = const,
	         var_ = flip ($),
		 sum_ = \bs st   -> sum $ map ($st) bs,
		 prod = \bs st   -> product $ map ($st) bs,
		 sub  = \b b' st -> b st-b' st,
		 div_ = \b b' st -> b st`div`b' st,
		 scal = \i b st  -> i*b st,
		 expo = \b i st  -> b st^i}	

evalRAlg :: Arith x (Store x -> Int)  	       -- foldArith evalRAlg = evalR
evalRAlg = Arith {con  = const,
	          var_ = flip ($), 
		  sum_ = fmap sum . sequence,
		  prod = fmap product . sequence,
		  sub  = liftM2 (-),
		  div_ = liftM2 div,
		  scal = \i   -> liftM2 (*) $ const i,
		  expo = \b i -> liftM2 (^) b $ const i}		   

evalWAlg :: Show x => Arith x (Store x -> (String,Int), Exp x)
					   -- foldArith evalWAlg = <writeVal,id>
evalWAlg = Arith {con  = \i   -> (const ("",i), Con i),
	          var_ = \x   -> let t = Var x in (\st -> out t $ st x, t),
		  sum_ = \fes -> let (fs,es) = unzip fes; t = Sum es
		                 in (\st -> do is <- mapM ($st) fs
		                 	       out t $ sum is, t),
		  prod = \fes -> let (fs,es) = unzip fes; t = Prod es
		                 in (\st -> do is <- mapM ($st) fs
		                 	       out t $ product is, t),
		  sub  = \(f,e) (g,e') -> let t = e :- e'
		     			  in (\st -> do i <- f st; k <- g st
		     			                out t $ i-k, t),
		  div_ = \(f,e) (g,e') -> let t = e :/ e'
		     			  in (\st -> do i <- f st; k <- g st
		     			                out t $ i`div`k, t),
		  scal = \i (f,e) -> let t = i :* e
		     	             in (\st -> do k <- f st; out t $ i*k, t),
		  expo = \(f,e) i -> let t = e :^ i
		     		     in (\st -> do k <- f st; out t $ k^i, t)}	
           where out e i = (show' e ++ " = " ++ show i ++ "\n",i)
	         show' = filter (/='\"') . show

evalWAlg' :: Show x => ArithP x (Store x -> (String,Int))
					       -- foldArithP evalWAlg = writeVal
evalWAlg' = ArithP {conP  = \i            -> const ("",i),
	            varP  = \x st        -> out (Var x) $ st x,
		    sumP  = \es bs st     -> fmap sum (mapM ($st) bs) >>= 
		   		             out (Sum es),
		    prodP = \es bs st     -> fmap product (mapM ($st) bs) >>= 
		   		             out (Prod es),
		    subP  = \e b e' b' st -> liftM2 (-) (b st) (b' st) >>= 
		   			     out (e :- e'),
		    divP  = \e b e' b' st -> liftM2 div (b st) (b' st) >>= 
		   			     out (e :/ e'),
		    scalP = \i e b st     -> fmap (i*) (b st) >>= out (i :* e),
		    expoP = \e b i st     -> fmap (^i) (b st) >>= out (e :^ i)}	
            where out e i = (show' e ++ " = " ++ show i ++ "\n",i)
	          show' = filter (/='\"') . show
	          
evalMAlg :: Arith x (Store x -> Maybe Int)         -- foldArith evalMAlg = evalM  
evalMAlg = Arith {con  = const . Just,
	          var_ = \x       -> Just . ($x),
	          sum_ = \bs st   -> do is <- mapM ($st) bs; Just $ sum is,
	          prod = \bs st   -> do is <- mapM ($st) bs; Just $ product is,
	          sub  = \b b' st -> do i <- b st; k <- b' st; Just $ i-k,
	          div_ = \b b' st -> do i <- b st; k <- b' st; guard $ k /= 0
	          		        Just $ i`div`k,
                  scal = \i b st  -> do k <- b st; Just $ i*k,
                  expo = \b i st  -> do k <- b st; Just $ k^i}
{-     
evalMAlg = Arith {con  = const . Just,
	          var_ = \x       -> Just . ($x),
	          sum_ = \bs st   -> fmap sum $ mapM ($st) bs,
	          prod = \bs st   -> fmap product $ mapM ($st) bs,
	          sub  = \b b' st -> liftM2 (-) (b st) $ b' st,
	          div_ = \b b' st -> do let m = b' st
	                                k <- m; guard $ k /= 0
	                                liftM2 div (b st) m,
                  scal = \i b st  -> fmap (i*) $ b st,
	          expo = \b i st  -> fmap (^i) $ b st}
-}

evalIOAlg :: Arith String (String -> MaybeT IO Int)	
					         -- foldArith evalIOAlg = evalIO
evalIOAlg = Arith {con  =  const . return,
	           var_ = \x file -> do store <- lift $ readFileContinue [] file 
	            			              $ return . read
	            	                let val = lookup x store
	            	                put $ x ++ case val of 
	            	                               Just i -> " = " ++ show i
	            	                               _ -> " is undefined"
	            	                liftT val,
	           sum_ = \bs file   -> do is <- mapM ($file) bs
	           			   return $ sum is,
	           prod = \bs file   -> do is <- mapM ($file) bs
	           			   return $ product is,
	           sub  = \b b' file -> do i <- b file; k <- b' file
	           			   return $ i-k,
	           div_ = \b b' file -> do i <- b file; k <- b' file
	                                   if k /= 0 then return $ i`div`k
	       		                   else do put "division by 0"
	       		                           mzero,
	           scal = \i b file  -> do k <- b file; return $ i*k,
	           expo = \b i file  -> do k <- b file; return $ k^i}
	    where put = lift . putStrLn 

evalIOWAlg :: Arith String (String -> MaybeT IO Int, Exp String)	
evalIOWAlg = Arith {con  = \i -> (const $ return i, Con i),
	            var_ = \x -> let t = Var x 
	                         in (\file 
	               -> do store <- lift $ readFileContinue [] file 
	               			   $ return . read
	            	     let val = lookup x store
	            	     case val of 
	            	          Just i -> out t i 
	            	          _ -> do lift $ putStrLn $ x ++ " is undefined"
	            	                  liftT Nothing, t),
	            sum_ = \fes 
	                      -> let (fs,es) = unzip fes; t = Sum es
	                         in (\file -> do i <- fmap sum $ mapM ($file) fs
	           			         out t i, t),
	            prod = \fes 
	                  -> let (fs,es) = unzip fes; t = Prod es
	                     in (\file -> do i <- fmap product $ mapM ($file) fs
	           			     out t i, t),
	            sub  = \(f,e) (g,e')
	                  -> let t = e :- e'
	                     in (\file -> do i <- liftM2 (-) (f file) $ g file
	           			     out (e :- e') i, t),
	            div_ = \(f,e) (g,e')
	                  -> let t = e :/ e'
	                     in (\file 
	                      -> do let m = g file
	                            k <- m 
	                            if k /= 0 then do i <- liftM2 div (f file) m
	                                              out (e :/ e') i 
	       		            else do lift $ putStrLn $ show' t
	       		              			       ++" is undefined"
	       		                    mzero,
	       		         t),
	            scal = \i (f,e) -> let t = i :* e
	                               in (\file -> do k <- fmap (i*) $ f file
	           			               out t k, t),
	            expo = \(f,e) i -> let t = e :^ i
	                               in (\file -> do k <- fmap (^i) $ f file
	           			               out t k, t)}
             where out e i = do lift $ putStrLn $ show' e ++ " = " ++ show i
                                return i
	           show' = filter (/='\"') . show

evalIOWAlg' :: ArithP String (String -> MaybeT IO Int)	
evalIOWAlg' = ArithP {conP = const . return,
	             varP = \x file 
	               -> do store <- lift $ readFileContinue [] file 
	               			   $ return . read
	            	     let val = lookup x store
	            	     case val of 
	            	          Just i -> out (Var x) i 
	            	          _ -> do lift $ putStrLn $ x ++ " is undefined"
	            	                  liftT Nothing,
	            sumP = \es bs file -> do i <- fmap sum $ mapM ($file) bs
	           			     out (Sum es) i,
	            prodP = \es bs file 
	                               -> do i <- fmap product $ mapM ($file) bs
	           			     out (Prod es) i,
	            subP = \e b e' b' file 
	                               -> do i <- liftM2 (-) (b file) $ b' file
	           			     out (e :- e') i,
	            divP = \e b e' b' file 
	                      -> do let m = b' file
	                            k <- m 
	                            if k /= 0 then do i <- liftM2 div (b file) m
	                                              out (e :/ e') i 
	       		            else do lift $ putStrLn $ show' (e :/ e')
	       		              			       ++" is undefined"
	       		                    mzero,
	           scalP = \i e b file -> do k <- fmap (i*) $ b file
	           			     out (i :* e) k,
	           expoP = \e b i file -> do k <- fmap (^i) $ b file
	           	 		     out (e :^ i) k}
             where out e i = do lift $ putStrLn $ show' e ++ " = " ++ show i
                                return i
	           show' = filter (/='\"') . show

diffE :: Eq x => Exp x -> x -> Exp x           -- symbolic differentiation
diffE e x = case e of Con _   -> zero
	              Var y   -> if x == y then one else zero
	              Sum es  -> Sum $ map d es
	              Prod es -> Sum $ zipWith h [0..] es
		                 where h i = Prod . updList es i . d
	              e :- e' -> d e :- d e'
	              e :/ e' -> (Prod [d e,e'] :- Prod [e,d e']) :/ (e':^2)
	              i :* e  -> i :* d e
	              e :^ i  -> i :* Prod [d e,e:^(i-1)]
            where d = flip diffE x
            
diffAlg :: Eq x => Arith x (x -> Exp x,Exp x)  -- foldArith diffAlg = <diffE,id>
diffAlg = Arith {con  = \i -> (const zero, Con i),
		 var_ = \x -> (\y -> if x == y then one else zero, Var x),
		 sum_ = \fes -> let (fs,es) = unzip fes
		                in (\x -> Sum $ map ($x) fs, Sum es),
		 prod = \fes -> let (fs,es) = unzip fes
		                    h x i e = Prod $ updList es i $ (fs!!i) x
		                in (\x -> Sum $ zipWith (h x) [0..] es, 
		                    Prod es),
		 sub  = \(f,e) (g,e') -> (\x -> f x :- g x, e :- e'),
		 div_ = \(f,e) (g,e') -> (\x -> (Prod [f x,e'] :- Prod [e,g x])
		                                 :/ (e' :^ 2),
		 			  e :/ e'),
		 scal = \i (f,e) -> (\x -> i :* f x, i :* e),
		 expo = \(f,e) i -> (\x -> i :* Prod [f x,e:^(i-1)], e :^ i)}

diffAlg' :: Eq x => ArithP x (x -> Exp x)      -- foldArithP diffAlg' = diffE
diffAlg' = ArithP {conP  = \_ _        -> zero,
		   varP  = \x y        -> if x == y then one else zero,
		   sumP  = \_ fs x     -> Sum $ map ($x) fs,
		   prodP = \es fs x    -> let f i e = Prod $ updList es i 
		 					   $ (fs!!i) x
		 			  in Sum $ zipWith f [0..] es,
		   subP  = \_ f _ g x  -> f x :- g x,
		   divP  = \e f e' g x -> (Prod[f x,e']:-Prod[e,g x]):/(e':^2),
		   scalP = \i _ f x    -> i :* f x,
		   expoP = \e f i x    -> i :* Prod [f x,e:^(i-1)]}

-- translation of arithmetic expressions into assembler programs
	 
data StackCom x = Push Int | Load x | Add Int | Mul Int | Sub | Div | Up
	          deriving Show

type Estate x = ([Int],Store x)

executeCom :: StackCom x -> Estate x -> Estate x
executeCom (Push a) (stack,store) = (a:stack,store)
executeCom (Load x) (stack,store) = (store x:stack,store) 
executeCom (Add n) st             = executeOp sum n st
executeCom (Mul n) st             = executeOp product n st
executeCom Sub st                 = executeOp (foldl1 (-)) 2 st
executeCom Div st                 = executeOp (foldl1 div) 2 st
executeCom Up st                  = executeOp (foldl1 (^)) 2 st
						
executeOp :: ([Int] -> Int) -> Int -> Estate x -> Estate x
executeOp f n (stack,store) = (f (reverse as):bs,store)
                              where (as,bs) = splitAt n stack

execute :: [StackCom x] -> Estate x -> Estate x
execute = flip $ foldl $ flip executeCom

codeAlg :: Arith x [StackCom x]
codeAlg = Arith {con  = \i    -> [Push i],
		 var_ = \x    -> [Load x],
		 sum_ = \cs   -> concat cs++[Add $ length cs],
		 prod = \cs   -> concat cs++[Mul $ length cs],
		 sub  = \c c' -> c++c'++[Sub],
		 div_ = \c c' -> c++c'++[Div],
		 scal = \i c  -> Push i:c++[Mul 2],
		 expo = \c i  -> c++[Push i,Up]}

code :: Exp x -> [StackCom x]				-- = foldArith codeAlg
code = \case Con i   -> [Push i]
	     Var x   -> [Load x]
	     Sum es  -> concatMap code es++[Add $ length es]
	     Prod es -> concatMap code es++[Mul $ length es]
	     e :- e' -> code e++code e'++[Sub]
	     e :/ e' -> code e++code e'++[Div]
	     i :* e  -> Push i:code e++[Mul 2]
	     e :^ i  -> code e++[Push i,Up]
                        
treeAlg :: Show x => Arith x (Tree String)		
treeAlg = Arith {con  = \i    -> int i,
		 var_ = \x    -> V $ show x,
		 sum_ = \ts   -> F "Sum" ts,
		 prod = \ts   -> F "Prod" ts,
		 sub  = \t t' -> F "-" [t,t'],
		 div_ = \t t' -> F "/" [t,t'],
		 scal = \i t  -> F "*" [int i,t],
		 expo = \t i  -> F "^" [t,int i]}	
	  where int i = F (show i) []

stringTree :: Show x => Exp x -> Tree String   		-- = foldArith treeAlg
stringTree = \case Con i   -> int i
	           Var x   -> V $ show x
	           Sum es  -> F "Sum" $ map stringTree es
	           Prod es -> F "Prod" $ map stringTree es
	           e :- e' -> F "-" [stringTree e,stringTree e']
	           e :/ e' -> F "/" [stringTree e,stringTree e']
	           i :* e  -> F "*" [int i,stringTree e]
	           e :^ i  -> F "^" [stringTree e, int i]
	     where int i = F (show i) []

storeAlg :: Read x => Store x -> TreeSig String Int
storeAlg store = TS {var = store . read, 
		     fun = \case "Sum" -> sum 
		       	         "Prod" -> product
		       	         "-" -> \[i,k] -> i-k
		       	         "/" -> \[i,k] -> i`div`k
		       	         "*" -> \[i,k] -> i*k
		       	         "^" -> \[i,k] -> i^k
		       	         i -> const $ read i}
	     
evalTree :: (Read x,Show x) => Exp x -> Store x -> Int 	-- = foldArith evalAlg
evalTree = flip (foldTree . storeAlg) . stringTree 

-- output of arithmetic expressions

instance Show x => Show (Exp x) where 
                   showsPrec _ (Con i)   = shows i
                   showsPrec _ (Var x)   = shows x 
                   showsPrec p (Sum es)  = enclose p q $ showMore q '+' es 
                           		   where q = prio '+'
                   showsPrec p (Prod es) = enclose p q $ showMore q '*' es 
                                           where q = prio '*'
                   showsPrec p (e :- e') = enclose p q $ showsPrec q e . ('-':) 
                   				       . showsPrec (q+1) e' 
                   		           where q = prio '-'  
                   showsPrec p (e :/ e') = enclose p q $ showsPrec q e . ('/':) 
                   				       . showsPrec (q+1) e' 
                   		           where q = prio '/'  
                   showsPrec p (i :* e)  = enclose p q $ shows i . ('*':) 
                                                       . showsPrec q e 
                                           where q = prio '*'
                   showsPrec p (e :^ i)  = enclose p q $ showsPrec q e . ('^':) 
                                                       . shows i 
                                           where q = prio '^'
		
prio :: Char -> Int
prio = \case '+' -> 0; '-' -> 0; '*' -> 1; '/' -> 1; '^' -> 2

enclose :: Int -> Int -> (String -> String) -> String -> String
enclose p q f | p > q = ('(':) . f . (')':) 
              | True  = f

showMore :: Show x => Int -> Char -> [Exp x] -> String -> String
showMore p op (e:es) = foldl f (showsPrec p e) es 
                       where f g e = g . (op:) . showsPrec p e    	           

-- reduction of arithmetic expressions

type Rstate x = (Int,[Exp x],Exp x -> Int)

updState :: Eq x => Rstate x -> Exp x -> Int -> Rstate x
updState (c,bases,f) e i = (c,insert e bases,update f e $ f e+i)

applyL :: ([a] -> a) -> [a] -> a
applyL _ [a] = a
applyL f as  = f as

reduceE :: Eq x => Exp x -> Exp x
reduceE = \case e :- e'       -> reduceE e :- reduceE e'
		i :* Con j    -> Con $ i*j
		0 :* e        -> zero
		1 :* e        -> reduceE e
		i :* (j :* e) -> reduceE $ (i*j) :* e
		i :* e        -> i :* reduceE e
		e :/ e'       -> reduceE e :/ reduceE e'
		Con i :^ j    -> Con $ i^j
		e :^ 0        -> one
		e :^ 1 	      -> reduceE e
		(e :^ i) :^ j -> reduceE $ e :^ (i*j)
		e :^ i        -> reduceE e :^ i
		Sum es        -> case f es of (c,[]) -> Con c
				              (0,es) -> applyL Sum es
				              (c,es) -> applyL Sum $ Con c:es
		Prod es       -> case g es of (c,[]) -> Con c
		     		     	      (1,es) -> applyL Prod es
		     		     	      (c,es) -> c :* applyL Prod es
		e -> e
      where f es = (c,map summand bases) where
                   (c,bases,scal) = foldl trans (0,[],const 0) $ map reduceE es
                   summand e = if i == 1 then e else i :* e where i = scal e
                   trans state@(c,bases,scal) = \case Con 0 -> state
	     					      Con i -> (c+i,bases,scal)
	     					      i:*e -> updState state e i
	     					      e -> updState state e 1
            g es = (c,map factor bases) where
                   (c,bases,expo) = foldl trans (1,[],const 0) $ map reduceE es
                   factor e = if i == 1 then e else e :^ i where i = expo e
                   trans state@(c,bases,expo) = \case Con 1 -> state
	     					      Con i -> (c*i,bases,expo)
	     					      e:^i -> updState state e i
	     					      e -> updState state e 1
                           
-- expression compiler

compE :: Arith String val -> Compiler val
compE alg = summand >>= moreSummands where
          summand       = mplus scalar factor >>= moreFactors
          factor        = msum [fmap (con alg) tint >>= power,
                                fmap (var_ alg) tidentifier >>= power,
	                        do tchar '('; a <- compE alg; tchar ')'
	                           power a]
          moreSummands a = msum [tchar '-' >> fmap (sub alg a) summand
			                   >>= moreSummands,
			         some (tchar '+' >> summand)
			                   >>= moreSummands . sum_ alg . (a:),
			         return a]
          moreFactors a  = msum [do tchar '/' >> fmap (div_ alg a) factor
            			 	      >>= moreFactors,
			         some (tchar '*' >> mplus scalar factor)
		                              >>= moreFactors . prod alg . (a:),
			         return a]
          scalar  = do i <- tint
                       let lift = fmap $ scal alg i
                       msum [tchar '*' >>
                              msum [lift scalar,
                                    fmap (var_ alg) tidentifier >>= lift . power,
                                    do tchar '('; a <- compE alg; tchar ')'
                                       return $ scal alg i a],
                            power $ con alg i]
          power a = msum [tchar '^' >> fmap (expo alg a) tint,
	       	          return a]

-- full compiler  

exp2alg :: String -> Int -> IO ()
exp2alg file n = readFileContinue () file h where
  h str = fromJust $ lookup n $ zip [0..] 
                [act parser out,
                 act (fmap diffE parser) $ out . ($"x"),		    -- 1
                 act (fmap (fst . foldArith diffAlg) parser) $ out . ($"x"),
                 act (fmap (snd . foldArith diffAlg) parser) out,
                 act (fmap (foldArithP diffAlg') parser) $ out . ($"x"),
                 act (compE $ mkArith diffAlg') $ out . ($"x") . snd,       -- 5
                 act (fmap eval parser) loopR,				    
           	 act (compE evalAlg) loopR,
           	 act (fmap evalM parser) loopR,
           	 act (compE evalMAlg) loopR,
           	 act (fmap evalTree parser) loopR,			    --10
           	 act (fmap evalR parser) loopR,				    
           	 act (compE evalRAlg) loopR,
           	 act (fmap code parser) loopC,
                 act (compE codeAlg) loopC,
                 act (fmap writeVal parser) $ loopT result,		    --15
                 act (fmap (fst . foldArith evalWAlg) parser) $ loopT result,
                 act (fmap (foldArithP evalWAlg') parser) $ loopT result,   
           	 act (compE $ mkArith evalWAlg') $ loopT result . snd,
           	 act (fmap evalIO parser) getStoreAndRun,
           	 act (compE evalIOAlg) getStoreAndRun,			    --20
           	 act (fmap (fst . foldArith evalIOWAlg) parser) getStoreAndRun, 
           	 act (fmap (foldArithP evalIOWAlg') parser) getStoreAndRun, 
           	 act (compE $ mkArith evalIOWAlg') $ getStoreAndRun . snd]   
          where parser :: Compiler (Exp String)
                parser = compE arithT
                out :: Exp String -> IO ()
                out = mkTree result . show . reduceE 
                result = file ++ "Result"
                act :: Compiler a -> (a -> IO ()) -> IO ()
           	act comp continue = case runST comp str of 
           	                    Just (a,"") -> continue a
				    Just (a,str) 
				      -> do putStrLn $ "unparsed suffix: "++str
			                    continue a
		                    _ -> putStrLn "syntax error"
			   
input :: IO (Store String,Bool)
input = do putStrLn "Enter variables!"; str <- getLine
           let vars = words str
	   putStrLn "Enter values!"; str <- getLine
	   let vals = map read $ words str
	       vals' = vals++replicate (length vars-length vals) 0
	   return (fold2 update (const 0) vars vals', not $ null vars)

loopR :: Show x =>(Store String -> x) -> IO ()
loopR val = do (store,b) <- input
               when b $ do putStrLn $ "result = "++show (val store)
                           loopR val
	   
loopT :: String -> (Store String -> (String,Int)) -> IO ()
loopT file val = do (store,b) <- input
                    when b $ do writeFile file $ fst $ val store
                                loopT file val
	            	   
loopC :: [StackCom String] -> IO ()
loopC code = do writeFile "code" $ fold2 f "" [0..] code
                loop 
             where f str n c = str ++ '\n':replicate (5-length lab) ' ' ++ 
	   		       lab ++ ": "++ show c where lab = show n
	           loop = do (store,b) <- input
		             let (result:_,_) = execute code ([],store)
		             when b $ do putStrLn $ "result = "++show result
		                         loop 

getStoreAndRun :: (String -> MaybeT IO Int) -> IO ()
getStoreAndRun val = do putStrLn "Enter file with store!"
                        file <- getLine
                        when (not $ null file) $ do i <- runMT $ val file
                        			    putStrLn $ show i
                     			            getStoreAndRun val
		  
-- EXAMPLES

exp1  = "5*11+6*12+x*y*z"
exp2  = "11*x^3+5*x^2+16*x+33"
exp3  = "5/(4-x)"
exp4  = "5*(6*(11*(x+y+z)*14+(c+g+b)*22)+44*(gg+hh))"
exp6  = "11*x^3*x^4+33+3*x+5*x^2+16*x+(6*x^2+55)*5*x^2"
exp7  = "x^4+5*x^3+11*x^2+222"
exp8  = "5*6*7+x-5*2*3"
exp9  = "9*x^2+xx/x-x*5*z+yy+5*6*x+yy/5*6/5"

{- contents of "store1": [("x",4),("y",55)]
   contents of "store2": [("z",4),("y",55)]

   exp2alg "exp1" 0		   --> 5*11+6*12+x*y*z	
   exp2alg "exp2" 0		   --> 11*x^3+5*x^2+16*x+33 		
   exp2alg "exp2" 1..3  	   --> 33*x^2+10*x+16
   exp2alg "exp2" 4..10 /\ x = 5   --> {Just} 1613
   exp2alg "exp7" 4..10 /\ x = 4   --> {Just} 974
   exp2alg "exp8" 14..16 /\ x = 66 --> 6*7 = 42
   				       5*6*7 = 210
   				       x = 66
   				       5*6*7+x = 276
   				       2*3 = 6
   				       5*2*3 = 30
   				       5*6*7+x-5*2*3 = 246
   exp2alg "exp3" 19/20 store1     --> x = 4
   				       division by 0
   				       Nothing 
   exp2alg "exp3" 21..23 store1    --> x = 4
   				       4-x = 0
   				       5/(4-x) is undefined
   				       Nothing
   exp2alg "exp7" 15..18 store1    --> x = 4
   				       x^4 = 256
   				       x = 4
   				       x^3 = 64
   				       5*x^3 = 320
   				       x = 4
   				       x^2 = 16
   				       11*x^2 = 176
   				       x^4+5*x^3+11*x^2+222 = 974 
   exp2alg "exp7" 19/20 store1     --> x = 4
   				       x = 4
   				       x = 4
   				       Just 974 
   exp2alg "exp7" 21..23 store1    --> x = 4
   				       x = 4
   				       x = 4
   				       x^4 = 256
   				       x = 4
   				       x^3 = 64
   				       5*x^3 = 320
   				       x = 4
   				       x^2 = 16
   				       11*x^2 = 176
   				       x^4+5*x^3+11*x^2+222 = 974 
   				       Just 974 -}
x = Var "x"
y = Var "y"
z = Var "z"
				        
t1  = Sum [5:*Con 11,6:*Con 12,Prod[x,y,z]]        	 -- 5*11+6*12+x*y*z	
t2  = Sum [Prod [x,y,z], Con 127]			 -- x*y*z+127
t3  = Sum [11 :* (x :^ 3),5 :* (x :^ 2),16 :* x,Con 33]  -- 11*x^3+5*x^2+16*x+33
t4  = Sum [11 :* (x :^ 3),5 :* (x :^ 2),16 :* x,Con 33,22 :* (x :^ 2)] 
t5  = Sum [5 :* (x :^ 2),22 :* (x :^ 2)] 
t6  = Prod [x :^5,Prod [x :^5,x :^6]]
t7  = Sum [11:*(x:^3),5:*(x:^2),16:*x,Con 33,x:-(y:-z),
           Prod[x:^5,Sum[x:^5,x:^6]]]
t8  = Sum [11:*(x:^3),5:*(x:^2),16:*x,Con 33,Sum[x,y,z]:-z,
           Prod[x:^5,Sum[x:^5,x:^6]]]
t9  = Prod [x,Con 5,5:*Prod[x:-y,y,z]]
t10 = Sum [Var"x":^4,5:*(Var"x":^3),11:*(Var"x":^2),Con 222]	
							 -- x^4+5*x^3+11*x^2+222
t11 = t10 >>= \"x" -> 2:*Con 2 				 -- 5/(4-x)
t12 = Con 5:/(Con 4:-Var"x")

store1 = \case "x" -> 4; "y" -> 55

val0 = foldArith evalAlg t10 store1	-- > 974
val1 = eval t10 store1           	-- > 974
val2 = eval t11 $ const 0 		-- > 974
val3 = evalM t10 store1 		-- > Just 974
val4 = evalM t12 store1 		-- > Nothing
val5 = evalR t10 store1			-- > 974
val6 = foldArith evalAlg t10 store1     -- > 974
val7 = foldArith evalMAlg t10 store1    -- > Just 974
val8 = runMT $ evalIO t10 "store1" 	-- > x = 4
					--   x = 4
					--   x = 4
					--   Just 974
val9 = runMT $ evalIO t12 "store1"      -- > x = 4
				        --   e/0 is undefined
				        --   Nothing
val10 = runMT $ evalIO t10 "store2"	-- > x is undefined
					--   Nothing
val11 = runMT $ evalIO t12 "store2"	-- > x is undefined
					--   Nothing
val12 = runMT $ evalIO t10 "store3"	-- > store3 does not exist
					--   x is undefined
				        --   Nothing
 				      
-- STEPWISE REDUCTIONS

mkCon :: Int -> Tree String
mkCon i = F "Con" [mkInt i]

-- differentiation

diffStep :: Rules String
diffStep (F "$" [F "diff" [V x],e]) = 
         case e of F "Con" [t] | isJust i 
   			         -> Just $ mkCon 0 where i = parse int $ root t
                   F "Var" [V y] -> Just $ mkCon $ if x == y then 1 else 0
                   F "Sum" es    -> Just $ F "Sum" $ map diff es
                   F "Prod" es   -> Just $ F "Sum" $ zipWith f [0..] es
                  		    where f i = F "Prod" . updList es i . diff
                   F "-" [e,e']  -> Just $ F "-" [diff e,diff e']
                   F "*" [t,e] | isJust $ parse int $ root t
	                         -> Just $ F "*" [t,diff e]
	                            where i = parse int $ root t
                   F "^" [e,t] | isJust i 
		    	         -> Just $ F "*" [t,F "Prod" [diff e,e']]
		                    where i = parse int $ root t
				          e' = F "^" [e,mkInt $ fromJust i-1]
		   _ -> Nothing
         where diff e = F "$" [F "diff" [V x],e]
diffStep _ = Nothing  

-- simplifying

redStep :: Rules String
redStep = \case F "Sum" es | any (== zero) es
	          -> Just $ F "Sum" $ filter (/= zero) es where zero = mkCon 0
                F "Sum" [e] -> Just e
                F "Prod" es | any (== one) es 
                  -> Just $ F "Prod" $ filter (/= one) es where one = mkCon 1
		F "Prod" [e] -> Just e
                F "*" [t,e] | isJust i && fromJust i == 1 -> Just e 
                	      where i = parse int $ root t
                F "*" [t,F "*" [u,e]] | isJust i && isJust j 
                  -> Just $ F "*" [mkInt $ fromJust i*fromJust j,e]
		     where [i,j] = map (parse int . root) [t,u]
                F "^" [e,t] | isJust i && fromJust i == 1 -> Just e 
                	      where i = parse int $ root t
                F "^" [F "^" [e,t],u] | isJust i && isJust j
		  -> Just $ F "^" [e,mkInt $ fromJust i*fromJust j] 
		     where [i,j] = map (parse int . root) [t,u]
		_ -> Nothing

-- evaluation

evalStep :: Store String -> Rules String
evalStep st (F "eval" [e]) = 
            case e of F "Con" [t] | isJust $ parse int $ root t 
            			    -> Just t
	              F "Var" [V x] -> Just $ mkInt $ st x
	              F "Sum" es    -> Just $ F "+" $ map eval es
	              F "Prod" es   -> Just $ F "*" $ map eval es
	              F "-" [e,e']  -> Just $ F "-" [eval e,eval e']
	              F "/" [e,e']  -> Just $ F "/" [eval e,eval e']
	              F "*" [t,e] | isJust $ parse int $ root t
	                            -> Just $ F "*" [t,eval e]
	              F "^" [e,t] | isJust $ parse int $ root t 
	                            -> Just $ F "^" [eval e,t]
	              _ 	    -> Nothing
	    where eval e = F "eval" [e]
evalStep _ _ = Nothing

exp10 = "diff(x)(Sum(Var(x)^4,5*(Var(x)^3),11*(Var(x)^2),Con(222)))"
exp11 = "eval(Sum(Var(x)^4,5*(Var(x)^3),11*(Var(x)^2),Con 222))"
exp12 = "Sum((Var(x)^4)^4,5*(6*(Var(x)^3)),11*(1*(Var(x)^2)),Con(0),Con(222))"

reduce1 = reduce "exp10" "x" [diffStep] 
reduce2 = reduce "exp11" "x" [evalStep store1] 
reduce3 = reduce "exp12" "x" [redStep] 

