module Lsys where
import Painter hiding (dragon)
import System.Directory
import System.Process
import System.Random

-- September 5, 2023

mkPng :: String -> IO ()
mkPng file = do rawSystem "convert" [file,"-trim",new]
		removeFile file
             where new = take (length file-4) file ++ ".png"

{- 
An L-system (d,a,s,rs) of type Lindenmayer contains the distance d by which the
turtle moves forward, the angle a by which it rotates to the left (+) or to the
right (-)‚ the start string s and the rewriting rules rs. The latter are stored 
here as pairs each one consisting of the character to be replaced and the 
replacing string.

Besides the L-system, the function lindenmayer has as parameters a drawing mode 
(see Painter.pdf) and the number n of replacements to be performed in parallel. 

iterM translates strings over {A,B,F,J,+,-,*,[,]} into turtle actions (see 
painter.hs) and performs the iterations of applyRules. If several rules can be 
applied to the same character, selects one of them at random.
-}

type Lindenmayer = (Float, Float, String, [(Char,String)])

lindenmayer :: Int -> Int -> Lindenmayer -> IO Curve
lindenmayer mode n (d,angle,a,rs) = 
     do str <- iterM f n a
        return $ turtle green mode $ Turn 270:map t str
     where t :: Char -> Action
           t '+' = Turn $ -angle
           t '-' = Turn angle
           t '*' = Turn 60
           t '/' = Turn $ -60
           t '[' = Open green 1 False
           t ']' = Close
           t 'F' = Move d
           t 'A' = Move d
           t 'B' = Move d
           t 'J' = Jump d
   	   t _   = Skip
           f a = do strs <- mapM (applyRules rs) a
                    return $ concat strs

iterM :: Monad m => (a -> m a) -> Int -> a -> m a
iterM f 0 a = return a
iterM f n a = do a <- f a; iterM f (n-1) a

rollDice :: [a] -> IO a
rollDice s = do i <- getStdRandom $ randomR (0,length s-1)
		return $ s!!i

applyRules :: [(Char,String)] -> Char -> IO String
applyRules rs c = if null applicable then return [c]
	          else do (_,r) <- rollDice applicable; return r
		  where applicable = filter ((== c) . fst) rs

-- Examples 

dragon :: Int -> Lindenmayer
dragon 4 = (2,-90,"A",[('A',"A-B"),('B',"A-B")])
dragon 5 = (5,-90,"A",[('A',"A+B+"),('B',"-A-B")])
dragon 6 = (2,-90,"AAAA",[('A',"A-B"),('B',"A+B")])
dragon _ = (1,-90,"AAAAAA",[('A',"A*B"),('B',"A/B")])

curve,dust,dust2,dust3,fern',sierp,tree' :: Lindenmayer
sierp   = (3,60,"A",[('A',"B-A-B"),('B',"A+B+A")])
dust    = (3,90,"F-F-F-F",[('F',"FF-F--F-F")])
dust2   = (3,90,"F-F-F-F",[('F',"FF-F--FF-F")])
dust3   = (3,90,"F*F*F*F*F*F",[('F',"FF-F--F-F")])
tree'   = (5,10,"FFX",[('X',"[+++[FX]+FFX][FFFX][--FX]"),('F',"FF"),
		       ('X',"[---[FX]-FFX][FFFX][++FX]")])
fern'   = (5,25,"X",[('X',"F-[[X]+X]+F[+FX]-X"),('F',"FF")])
curve   = (44,24,"X",[('X',"F-X"),('X',"F+X")])

draw1,draw2 :: Int -> Lindenmayer -> IO ()
draw1 n system = do c <- lindenmayer 14111 n system
	            drawC (2,2) $ flipH c
draw2 n system = do c <- lindenmayer 12111 n system
	            drawC (2,2) $ flipH c
	            
draw3 :: Int -> Int -> Lindenmayer -> IO ()
draw3 m n system = do cs <- sequence $ replicate m 
				     $ lindenmayer 12111 n system
	              drawCS True "lind2" cs

