{-# LANGUAGE FlexibleInstances, LambdaCase, TypeFamilies, CPP #-}

module Painter where

------------------------------------ Copyright (c) Peter Padawitz, July 23, 2022

---- Manual: http://fldit-www.cs.uni-dortmund.de/~peter/Haskellprogs/Painter.pdf

import Prelude hiding (catch,showList)
import Control.Exception
import Control.Applicative(Applicative(pure,(<*>)),Alternative(empty,(<|>)))
import Control.Monad
import Data.Complex
import Data.Maybe (fromJust,isJust)
import System.Directory
import Data.Map.Strict (Map,(!),keys,notMember,filterWithKey,fromList,delete)
import qualified Data.Map.Strict as DMS (empty,lookup,insert,union,null)
import Movie(moviePix,movies,movieSelect,html,(&),embed,mkFile,search)

#if __GLASGOW_HASKELL__ >= 806
import Control.Monad.Fail hiding (fail)		
import qualified Control.Monad.Fail (fail)	
instance MonadFail Parser where fail = fail
instance MonadFail m => MonadFail (MaybeT m) where fail = fail
#endif

(<$>) :: Monad m => (a -> b) -> m a -> m b
(<$>) = liftM

(&&&),(|||) :: (a -> Bool) -> (a -> Bool) -> a -> Bool
(&&&) = liftM2 (&&)
(|||) = liftM2 (||)

add1 :: Num a => (a,a) -> a -> (a,a)
add1 (x,y) a = (x+a,y)
 
add2,sub2 :: Num a => (a,a) -> (a,a) -> (a,a)
add2 (x,y) (a,b) = (a+x,b+y)
sub2 (x,y) (a,b) = (a-x,b-y)
 
apply2 f (x,y) = (f x,f y)

float :: RealFloat a => Int -> a
float = fromInteger . toInteger
	 
lgW a = float $ 5*length a

float2 (x,y) = (float x,float y)

round2 :: (RealFrac b, Integral d, RealFrac a, Integral c) => (a,b) -> (c,d)
round2 (x,y) = (round x,round y)
 
-- minmax ps computes minimal and maximal coordinates of the point list ps.

minmax ps@((x,y):_) = foldl f (x,y,x,y) ps
             where f (x1,y1,x2,y2) (x,y) = (min x x1,min y y1,max x x2,max y y2)
minmax _ = (0,0,0,0)
		     
fixpt :: (a -> a -> Bool) -> (a -> a) -> a -> a
fixpt rel step a = if rel b a then a else fixpt rel step b where b = step a

infixr 2 `union`
infixr 3 `meet`

indices s = [0..length s-1]

context s i = take i s++drop (i+1) s

updList s i a = take i s++a:drop (i+1) s
 
update :: Eq a => (a -> b) -> a -> b -> a -> b
update f a b x = if x == a then b else f x

fold2 :: (a -> b -> c -> a) -> a -> [b] -> [c] -> a
fold2 f a (b:bs) (c:cs) = fold2 f (f a b c) bs cs
fold2 _ a _ _           = a
 
zipWith4 f (a:as) (b:bs) (c:cs) (d:ds) = f a b c d:zipWith4 f as bs cs ds
zipWith4 _ _ _ _ _ 		       = []

insert,remove :: Eq a => a -> [a] -> [a]
insert x s@(y:s') = if x == y then s else y:insert x s'
insert x _        = [x]
remove = filter . (/=)

union,diff,meet :: Eq a => [a] -> [a] -> [a]
union = foldl $ flip insert 
diff  = foldl $ flip remove
meet  = filter . flip elem

mkSet :: Eq a => [a] -> [a]
mkSet = union []

unionMap :: Eq b => (a -> [b]) -> [a] -> [b]
unionMap f = foldl union [] . map f

supset,subset,shares,disjoint :: Eq a => [a] -> [a] -> Bool
supset s   = all (`elem` s)  
subset     = flip supset
shares s   = any (`elem` s)
disjoint s = not . shares s

sort r (x:s) = sort r [y | y <- s, r y x] ++ x:sort r [y | y <- s, not $ r x y]
sort _ s     = s
	        		  
liftR :: (Eq a,Eq b) => [(a,b)] -> [a] -> [b] -> Bool
liftR rel as bs = all (\a -> any (\b -> (a,b) `elem` rel) bs) as &&
		  all (\b -> any (\a -> (a,b) `elem` rel) as) bs
 
kfold :: (a -> state -> [state]) -> [state] -> [a] -> [state]
kfold f = foldl $ \states a -> concatMap (f a) states               
 
ktest = length$kfold(\a st ->[st,st])[0][1..6]      	-- = 2^6       

newlines (str:strs@(_:_)) = str++",\n"++newlines strs
newlines strs             = concat strs

removeComment :: Int -> String -> String
removeComment 0 ('-':'-':str) = removeComment 1 str
removeComment 0 ('{':'-':str) = removeComment 2 str
removeComment 0 (x:str)       = x:removeComment 0 str
removeComment 1 ('\n':str)    = '\n':removeComment 0 str
removeComment 1 (_:str)       = removeComment 1 str
removeComment 2 ('-':'}':str) = removeComment 0 str
removeComment 2 (_:str)       = removeComment 2 str
removeComment _ _             = []

searchGet :: (a -> Bool) -> [a] -> Maybe (a,Int)
searchGet f = g 0 where g i (a:s) = if f a then Just (a,i) else g (i+1) s
                        g _ _     = Nothing

max2:: (Ord a,Num a) => [(a,a)] -> (a,a)
max2 = foldl1 f where f (x,y) (a,b) = (max x a,max y b)
                     
rel2fun :: Eq a => [(a,b)] -> a -> [b]
rel2fun = foldl f $ const [] where f g (a,b) = update g a $ b:g a
                     
fun2invrel :: (a -> [b]) -> [a] -> [(b,a)]
fun2invrel f as = [(b,a) | a <- as, b <- f a]
           				      		  
-- monad transformers

newtype MaybeT m a = MaybeT {runMT :: m (Maybe a)}

instance Monad m => Monad (MaybeT m) where
         return = MaybeT . return . Just
 	 m >>= f = MaybeT $ do ma <- runMT m
 			       case ma of Just a -> runMT $ f a
 			       		  _ -> return Nothing 	
 	 fail _ = MaybeT $ return Nothing

instance Monad m => MonadPlus (MaybeT m) where
 	 mzero = MaybeT $ return Nothing
         m `mplus` m' = MaybeT $ do ma <- runMT m
 			            if isJust ma then return ma else runMT m' 

instance Monad m => Alternative (MaybeT m) where empty = mzero
 	 					 (<|>) = mplus

instance Monad m => Functor (MaybeT m) where fmap = liftM
 	 
instance Monad m => Applicative (MaybeT m) where pure  = return         		       
		          		         (<*>) = liftM2 id
 	 					 
class MonadTrans t where type M t :: * -> *
			 lift  :: Monad m => m a -> t m a
			 liftT :: Monad m => M t a -> t m a			 
		        
instance MonadTrans MaybeT where 
	            type M MaybeT = Maybe 
	            lift  = MaybeT . liftM Just -- :: m a -> MaybeT m a
		    liftT = MaybeT . return     -- :: Maybe a -> MaybeT m a
		    
-- TYPES FOR PAINTING TREES AND PATHS

type Point  = (Float,Float)
type LPoint = (String,Float,Float)
type CPoint = (RGB,Float,Float)
type Path   = [Point]
type LPath  = [LPoint]
type CPath  = (RGB,Int,Path)
type CLPath = (RGB,LPath)

-- READ, WRITE AND DRAW
				    
test :: (Read a,Show b) => (a -> b) -> IO ()
test f = readFileContinue () "source" $ writeFile "target" . show . f . read

readFileContinue :: a -> String -> (String -> IO a) -> IO a
readFileContinue err file act = do str <- readFile file `catch` handler
		                   if null str 
		                      then do putStrLn $ file++" does not exist"
		            		      return err
			              else act str
			        where handler :: IOError -> IO String
			      	      handler _ = return ""
			      
type Svg a = Point -> a -> (String,(Int,Int))    -- for creating svg code from a
	
readAndDraw :: Point -> String -> String -> Svg a -> (String -> Maybe a) 
						  -> IO ()
readAndDraw sc file suffix draw parse = 
 readFileContinue () file $ drawSingle sc (file++suffix) draw . fromJust . parse

readAndDrawDir :: Show a => String -> Parser a -> (Int -> Svg a) -> IO ()
readAndDrawDir dir parser draw = readFileContinue () dir $ act
	           where act str = do mkHtml True dir n
			              drawDir p2 dir "" draw as n
				   where as = fromJust $ parse (more parser) str
	  			         n = length as

getScaleAndAct :: (Point -> String -> IO ()) -> IO () -> IO ()
getScaleAndAct act continue = do putStrLn prompt
    		                 str <- getLine
    		                 let strs = words str
    		                 when (length strs >= 2) $ 
		                   do let [hor,ver] = map read $ take 2 strs
			                  rest = drop 2 strs
			              act (hor,ver) $ 
			                  if null rest then "" else head rest
			              continue

prompt = "Enter both a horizontal and a vertical scaling factor and - " ++
   "optionally - a background image with suffix .svg, .jpg, .png, .gif or .pdf!"

-- drawSingle p0 file draw a loads a from file, enters a loop that asks for 
-- scaling factors and an optional background and draws a into Pix/file.svg.
			    	                      	
drawSingle :: Point -> String -> Svg a -> a -> IO ()
drawSingle sc file draw a = do b <- doesDirectoryExist "Pix"
			       when (not b) $ createDirectory "Pix"
			       scale
                         where scale = if sc == p0 then getScaleAndAct act scale 
        			                   else act sc ""
                               act sc back = writeFile ("Pix/" ++ file ++".svg")
			    	                       $ embed back $ draw sc a

-- For nonempty lists as, drawList dir back draw as creates the directory 
-- Pix/dir, draws as!!i for all i in indices(ts) with a horizontal resp. 
-- vertical scaling factor 10 resp. 40 into Pix/dir/i.svg and colors red resp. 
-- green all nodes of the subtree of ts!!i with root p if i is even resp. odd.

drawList :: String -> String -> (Int -> Svg a) -> [a] -> IO ()
drawList dir back draw as = do b <- doesDirectoryExist "Pix"
			       when (not b) $ createDirectory "Pix"
			       mkHtml True dir n
			       drawDir p2 dir back draw as n
			    where n = length as
 
mkHtml :: Bool -> String -> Int -> IO ()
mkHtml frame dir n = html frame ("Pix/" ++ dir) $ map (mkFile dir) [1..n]

drawDir :: Point -> String -> String -> (Int -> Svg a) -> [a] -> Int -> IO ()
drawDir sc dir back draw as n = do b <- doesDirectoryExist pdir
                   	           when b $ removeDirectoryRecursive pdir
                   	           createDirectory pdir
                   	           mapM_ write [1..n]
                      		   when (not $ null back) $
                      		        copyFile ("Pix/"++back) $ pdir++'/':back
                  where pdir = "Pix/" ++ dir
                        write i = writeFile ("Pix/" ++ mkFile dir i) 
                                            $ embed back $ draw i sc $ as!!(i-1)

-- drawText file loads a string from file and writes it in columns each of which
-- consists of at most 37 lines into file.svg.

drawText :: String -> IO ()
drawText file = do b <- doesDirectoryExist "Pix"
                   when (not b) $ createDirectory "Pix"
		   readFileContinue () file act 
	        where act str = writeFile ("Pix/" ++ file ++ ".svg")
	        			  $ embed "" (code,size)
		        where code = concatMap drawCol [0..nrCols-1]
			      size = (9*pos nrCols+45,20*maximum heights)
			      ls = lines str
			      nrLines = length ls
			      matrix = split ls
			      nrCols = length matrix
			      heights = map length matrix
			      split s = if length s <= 37 then [s]
			     	        else take 37 s:split (drop 37 s)
			      pos 0 = 1
			      pos i = pos k+maximum (map length $ matrix!!k)
		                      where k = i-1
		              drawCol i = concat $ zipWith (drawLine $ pos i) 
	 			                   [0..heights!!i-1] $ matrix!!i
			      drawLine p k = text black $ float2 (9*p,24+k*20)
	 
-- LABELLED TREES

data Tree a = V a | F a [Tree a] | E a (Tree a) deriving (Eq,Show)

-- getSub as t returns the subtree(s) of t at the end of the path as of edge 
-- labels.

getSub :: (Eq a,MonadPlus m) => [a] -> Tree a -> m (Tree a)
getSub [] t           = return t
getSub as (F _ ts)    = msum $ map (getSub as) ts
getSub (a:as) (E b t) = if a == b then getSub as t else mzero

-- putSub as t u returns the tree(s) that result(s) from t by replacing the 
-- tree(s) at the end of the path as of edge labels with u.

putSub :: (Eq a,MonadPlus m) => [a] -> Tree a -> Tree a -> m (Tree a)
putSub [] sub _           = return sub
putSub as sub (F x ts)    = msum $ map f $ indices ts
			    where f i = do t <- putSub as sub $ ts!!i
			 		   return $ F x $ updList ts i t
putSub (a:as) sub (E b t) = if a == b then do t <- putSub as sub t
					      return $ E b t 
			              else mzero

type Equations a = [(Tree a,Tree a)]
type Rules a     = Tree a -> Maybe (Tree a)

isV (V _) = True
isV _     = False

isLeaf (V _)    = True
isLeaf (F _ []) = True
isLeaf _        = False

root :: Tree a -> a
root (V a)   = a
root (F a _) = a
root (E a _) = a

subtrees :: Tree a -> [Tree a]
subtrees (F _ ts) = ts
subtrees (E _ t)  = [t]
subtrees _        = []

vars :: Eq a => Tree a -> [a]
vars (V a)    = [a]
vars (F _ ts) = unionMap vars ts
vars (E _ t)  = vars t

height :: Tree a -> Int
height (F _ ts) = foldl max 0 (map height ts)+1
height (E _ t)  = height t+1
height _        = 1

mapTree :: (a -> b) -> Tree a -> Tree b
mapTree f (V a)    = V $ f a
mapTree f (F a ts) = F (f a) $ map (mapTree f) ts
mapTree f (E a t ) = E (f a) $ mapTree f t

foldTree :: (a -> [b] -> b) -> Tree a -> b
foldTree f (V a)    = f a []
foldTree f (F a ts) = f a $ map (foldTree f) ts
foldTree f (E a t)  = f a [foldTree f t]
  
(>>>) :: Tree String -> (String -> Tree String) -> Tree String
V a >>> sub    = sub a  
F a ts >>> sub = F a $ if a == "&#955;" then f ts else map (>>> sub) ts
		 where f (F "->" [p,t]:ts)   = F "->" [p,t >>> sub' p]:f ts
		       f (F "->" [p,b,t]:ts) = F "->" [p,b,t >>> sub' p]:f ts
		       f ts                  = ts
		       sub' p x = if x `elem` vars p then V x else sub x

type Node = [Int]

(<<),(<<=) :: Node -> Node -> Bool
(m:p) << (n:q) = m == n && p << q
_ << p         = not $ null p
p <<= q        = p << q || p == q

nodes,leaves :: Tree a -> [Node]

nodes (F _ ts)  = []:[i:p | (t,i) <- zip ts [0..], p <- nodes t]
nodes (E _ t)   = []:[0:p | p <- nodes t]
nodes _         = [[]]

leaves (F _ []) = [[]]
leaves (F _ ts) = [i:p | (t,i) <- zip ts [0..], p <- leaves t]
leaves (E _ t)  = []:[0:p | p <- leaves t]
leaves _        = [[]]

label :: Tree a -> Node -> a
label t []           = root t
label (F _ ts) (n:p) = label (ts!!n) p
label (E _ t)  (_:p) = label t p

-- zipWithSucs f p s applies the binary function f to the pairs obtained by 
-- zipping the list of direct successor positions of p with s.

zipWithSucs :: (Node -> a -> b) -> Node -> [a] -> [b]
zipWithSucs f p = zipWith f $ map (\i -> p++[i]) [0..]

class HasVars a where isVar :: a -> Bool; getVar :: a -> a

instance HasVars Int where isVar = (<0); getVar x = -x
instance HasVars String where isVar str = not (null str) && head str == 'V'
			      getVar (_:str) = str

readTree :: (Read a,HasVars a) => ReadS (Tree a)
readTree s = forks++vars++consts
	     where vars   = [(V $ getVar a,s1) | (a,s1) <- reads s, isVar a]
	           consts = [(F a [],s1) | (a,s1) <- reads s, not $ isVar a]
		   forks  = [(F a $ t:ts,s4) | (a,s1) <- reads s,
		                               ("(",s2) <- lex s1,
					       (t,s3) <- readTree s2,
                                               (ts,s4) <- trees s3]
		   trees s = [([],s1) | (")",s1) <- lex s] ++
		   	     [(t:ts,s3) | (",",s1) <- lex s,
		   		          (t,s2) <- readTree s1,
					  (ts,s3) <- trees s2]

instance (Read a,HasVars a) => Read (Tree a) where readsPrec _ = readTree

tree1 = read "5(66,77(8,-5))" :: Tree Int
tree2 = read "\"5\"(\"66\",\"77\"(\"8\",\"V3\"))" :: Tree String

-- PARSER

data Parser a = P {runP :: String -> Maybe (a,String)}

instance Monad Parser where
         return a = P $ \str -> Just (a,str)
         P h >>= f = P $ (>>= \(a,st) -> runP (f a) st) . h
	 fail _ = mzero

instance MonadPlus Parser where mzero = P $ const Nothing
         		        P g `mplus` P h = P $ liftM2 mplus g h

{- enforced constraints: Functor f => Applicative f
   			 Applicative f => Alternative f
   			 Applicative m => Monad m
   			 (Alternative m, Monad m) => MonadPlus m -}

instance Functor Parser where fmap f m = do x <- m; return $ f x

instance Applicative Parser where pure  = return
          			  (<*>) = liftM2 id

instance Alternative Parser where empty = mzero
 	 			  (<|>) = mplus

parse :: Parser a -> String -> Maybe a
parse p str = do (a,_) <- runP p str; Just a

parseBool :: Tree String -> Maybe Bool
parseBool = parse bool . root

parseInt :: Tree String -> Maybe Int
parseInt = parse int . root
 
sat :: Parser a -> (a -> Bool) -> Parser a
sat p f = do a <- p; guard $ f a; return a

item = P $ \str -> do c:str <- Just str; Just (c,str)

some,many,nelist,list :: Parser a -> Parser [a]
some p   = do a <- p; fmap (a:) $ many p
many p   = some p `mplus` return []
more p = do a <- p; msum [do tchar ','; as <- more p; return $ a:as, 
			  return [a]]
nelist p = p >>= \a -> msum [tchar ',' >> fmap (a:) (nelist p), 
			     return [a]]
list p   = tchar '[' >> msum [tchar ']' >> return [],
	                      do as <- nelist p; tchar ']'; return as]

noSpecial = sat item (`notElem` " \n\t()[]{},;<>=/\\|!&#:.$+-*^") 

char,nchar,tchar :: Char -> Parser Char
char x  = sat item (== x)
nchar x = sat item (/= x)
tchar = token . char

digit,letter,delim :: Parser Char
digit  = msum $ map char ['0'..'9']
letter = msum $ map char $ ['a'..'z']++['A'..'Z']
delim  = msum $ map char " \n\t"

infixes :: [String]
infixes = words 
	  ";; ; >> = ==> \\/ /\\ == /= <= < >= > $ # : + - * / ^^ ^ . !! &"

string,tstring :: String -> Parser String
string  = mapM char
tstring = token . string

identifier :: Parser String 
identifier = liftM2 (:) letter $ many noSpecial

nat,int :: Parser Int
nat = fmap read $ some digit
int = nat `mplus` do tchar '('; tchar '-'; n <- nat; tchar ')'; return $ -n

real :: Parser Float
real = do i <- int; char '.'; ds <- some digit; let n = read ds
	  return $ fromIntegral i+fromInteger n/fromIntegral (10^length ds)

quoted,getOp :: Parser String
quoted = char '"' >> do x <- many $ nchar '"'; char '"'; return x
getOp  = msum $ map tstring infixes

function :: Parser String -> Parser String
function var = token $ msum [quoted, var >> mzero, some noSpecial]

bool :: Parser Bool
bool = msum [string "true"  >> return True,
             string "false" >> return False]
             
relation :: Ord a => Parser (a -> a -> Bool)
relation = msum [string "==" >> return (==),
                 string "/=" >> return (/=),
                 string "<=" >> return (<=),
                 string ">=" >> return (>=),
                 string "<"  >> return (<),
                 string ">"  >> return (>)]

token :: Parser a -> Parser a
token p = do many delim; a <- p; many delim; return a

termEqs :: [String] -> Parser (Tree String,[Equations String])
termEqs xs = do t <- term $ sat identifier (`elem` xs)
--termEqs xs = do t <- term $ msum $ map string xs
	        return $ case t of F ";;" (t:ts) -> (t,map eqs ts)
                                   _ -> (t,[])
	     where eqs (F ";" ts)    = [(t,u) | F "=" [t,u] <- ts]
	           eqs (F "=" [t,u]) = [(t,u)]
		   eqs _             = []

termAndNode :: Parser (Tree String,Node)
termAndNode = do tchar '('; t <- term mzero; tchar ','; node <- list nat
		 tchar ')'; return (t,node)

term :: Parser String -> Parser (Tree String)
term var = term' infixes where

     term' :: [String] -> Parser (Tree String)
     term' (a:as) = do t <- term' as
		       ts <- many $ tstring a >> term' as
 	               return $ if null ts then t else F a $ t:ts
     term' _      = msum [tree,
                          token var >>= curryrest . V,
                          tuple >>= curryrest,
                          termList >>= curryrest,
     			  do r <- token real; curryrest $ mkReal r,
                          do i <- token int; curryrest $ mkInt i,
                          do op <- function var; curryrest $ F op []]
                             
     tree,tuple,abstraction,termList :: Parser (Tree String)
     tree = msum [tchar 'V' >> fmap V (token quoted),
                  tchar 'F' >> do x <- token quoted; F _ ts <- termList
                                  return $ F x ts,
                  tchar 'E' >> do x <- token quoted; t <- term var
                                  return $ E x t]
     tuple = do tchar '('; msum [tchar ')' >> return (F "()" []),
 			         do abs <- abstraction
 			            abss <- many $ do tchar ','; abstraction
 			            tchar ')'; return $ F "&#955;" $ abs:abss,
 			         do op <- getOp; tchar ')'; return $ F op [],
 			         do op <- getOp; t <- term var; tchar ')'
 			            return $ F "rsec" [F op [],t], 
 			         do t <- term var; op <- getOp; tchar ')'
 			            return $ F "lsec" [t,F op []],
	      		         do ts <- nelist $ term var; tchar ')'
			            return $ case ts of [t] -> t
			            		        _ -> F "()" ts]
     abstraction = do p <- term var
     		      msum [tstring "->" >> do t <- term var
		                               return $ F "->" [p,t],
		            tstring "|" >> do b <- term var; tstring "->"
		                              t <- term var
		                              return $ F "->" [p,b,t]]
     termList = msum [tchar '[' >> do t <- term var; tstring ".."
     			              u <- term var; tchar ']'
	                              return $ F "range" [t,u],
	              do ts <- list $ term var; return $ F "[]" ts]
	              
     fields :: Parser [Tree String]
     fields = do field <- function var; tchar '='; t <- term var
	         let val = E field t
	         msum [tchar ',' >> fmap (val:) fields,
	               return [val]]
	              
     curryrest :: Tree String -> Parser (Tree String)
     curryrest t = msum [tuple >>= more,
		         termList >>= more,
		         do tchar '{'; ts <- fields; tchar '}'
		            F op [] <- return t; return $ F op ts,
		         --do tchar '('; op <- function var; tchar ')'; more $ F op [],
		         --term var >>= more,
		         return t] 
	  where more u = curryrest $ case t of 
	                             F x [] | root u == "()" -> F x $ subtrees u 
	    			            | True           -> F x [u] 
                                     _ -> F "$" [t,u]
	
termPair :: Parser (Tree String,Tree String)
termPair = do tchar '('; t <- term mzero; tchar ','; t' <- term mzero; tchar ')'
	      return (t,t')
      
termNodes :: Parser (Tree String,[Node])
termNodes = do t <- term mzero; tchar ','; ps <- list $ list nat; return (t,ps)

-- TREE PAINTER

mkTree file str  = do writeFile file str; drawTerm file
mkTreeC file str = do writeFile file str; drawTermC file
mkTreeN file str = do writeFile file str; drawTermNodes file
mkTrees file str = do writeFile file str; drawTerms file ""

tree3  = mkTree "tree3" "32(25(17,30 27,zzz h),50(42,63 89))"
trees1 = mkTrees "trees1" $ "32 [25 [17,30 [27]],50 [42,zz%hhh [89]]]32," ++
			    "[25 [17,30 [27]],50 [42,63 [89]]]"
treeN1 = mkTreeN "treeN1" "32(25(17,30 27,zzz h),50(42,63 89)),[[0,1,0],[0,2]]"

-- drawTerm{C} file reads the contents of file, compiles it into a tree t of
-- type Tree String and draws t with a horizontal resp. vertical scaling factor
-- 10 resp. 40 into file.svg.
-- drawTerm{C}I asks for the scaling factors and an optional background before
-- drawing t.
-- drawTermC colors the nodes at level i with hue 1 red (height t) i.
-- drawTermNodes also expects a list of nodes that will be colored green.			    	                    
-- drawTerms{C} dir reads a nonempty list ts of trees from file, creates the 
-- directory Pix/dir and draws ts!!i for all i in indices(ts) with a horizontal
-- resp. vertical scaling factor 10 resp. 40 into Pix/file/i.svg.
	
drawTerm,drawTermC,drawTermNodes,drawTermsP :: String -> IO ()
drawTerm file      = readAndDraw p0 file "" svgTree $ parse $ term mzero
drawTermC file     = readAndDraw p0 file "" svgTreeC $ parse $ term mzero
drawTermNodes file = readAndDraw p0 file "" svgTreeSet $ parse termNodes
drawTermsP file    = readAndDrawDir file termAndNode svgTreeP  -- for Tableau.hs

-- drawTerm{C}I file f reads file, transforms the contents of file into a tree 
-- t, applies f to t computes SVG code from the resulting tree and saves the 
-- corresponding picture in Pix/fileResult.svg.
-- drawTermI2 file f works analogously to drawTermI for binary functions f, 
-- i.e., file is supposed to contain a pair of terms.

drawTermF,drawTermCF :: String -> Rules String -> IO ()
drawTermF file f  = readAndDraw p0 file "Result" g $ parse $ term mzero 
                    where g sc t = svgTree sc $ maybeError $ f t
drawTermCF file f = readAndDraw p0 file "Result" g $ parse $ term mzero 
                    where g sc t = svgTreeC sc $ maybeError $ f t

drawTermF2 :: String -> (Tree String -> Rules String) -> IO ()
drawTermF2 file f = readAndDraw p0 file "Result" g $ parse termPair 
                    where g sc (t,t') = svgTree sc $ maybeError $ f t t'

maybeError :: Maybe (Tree String) -> Tree String
maybeError (Just t) = t
maybeError _        = F "error" []
                   		                           
-- Sample calls: drawTermI "baum1" $ getSub $ words "e3 e4 e1"
--		 drawTermI "baum1" $ getSub $ words "e3 e4 e22"
--		 embedFun2 "baumpaar" $ putSub $ words "e3 e4 e1 e4"
--		 embedFun2 "baumpaar" $ putSub $ words "e3 e4 e22"

drawTerms,drawTermsC :: String -> String -> IO ()
drawTerms dir back  = readFileContinue () dir $ drawList dir back 
			(const svgTree) . fromJust . parse (nelist $ term mzero)
drawTermsC dir back = readFileContinue () dir $ drawList dir back 
		       (const svgTreeC) . fromJust . parse (nelist $ term mzero)

-- Given an XCTL formula form of type nodeSet (see above), drawNodeSet file form 
-- reads a tree t from file, draws t with a horizontal resp. vertical scaling 
-- factor 10 resp. 40 into file_set.svg and colors green all nodes of form(t).

drawNodeSet :: String -> String -> IO ()
drawNodeSet file form = readFileContinue () file act where
       act tstr = case phi of Just set -> drawSingle p2 file' svgTreeSet (t,set)
	         	      _ -> drawSingle p2 file' svgTree $ F "NO NODES" []
	          where t = fromJust $ parse (term mzero) tstr
	                phi = (parse (term mzero) >=> foldSet (xalg t)) form
	                file' = file ++ "_set"
	         	           	    
-- Given an XCTL formula form of type nodeRel (see above), drawNodeRel file form
-- reads a tree t from file, creates the directory Pix/file_rel, draws t with a 
-- horizontal resp. vertical scaling factor 10 resp. 40 for 
-- ws = [w in nodes(t) | form(t)(w)/=[]] and all i in indices(ws) into 
-- Pix/file_rel/i.svg  and colors ws!!i red and green all nodes of
-- form(t)(ws!!i).

drawNodeRel :: String -> String -> IO ()
drawNodeRel dir form = readFileContinue () dir act where
	 act tstr = case phi of Just rel -> drawL $ trips rel
	         	        _ ->  draw $ F "NO NODEREL" []
	            where t = fromJust $ parse (term mzero) tstr
	                  phi = (parse (term mzero) >=> foldRel (xalg t)) form
	                  trips rel = [(t,w,ws) | w <- nodes t, 
	                      	                  let ws = rel w, not $ null ws]
	             	  draw = drawSingle p2 (dir ++ "_set") svgTree
	             	  drawL = drawList (dir ++ "_rel") "" $ const svgTreeRel
                  
-- REDUCER

reduce = reduceTerm 100

-- reduceTerm n file vars rules subsequently applies to the term in file the 
-- equations in file, then rules!!0, then rules!!1, etc., and dyes redices red 
-- and reducts green. vars yields the symbols that the parser turns into 
-- variables. The reduction stops after n steps.

reduceTerm :: Int -> String -> String -> [Rules String] -> IO ()
reduceTerm n file vars rules = readFileContinue () file f where
             dir = file ++ "Reduction"
	     f str = case parse (termEqs $ words vars) $ removeComment 0 str of 
	  	          Just (t,eqs) -> drawList dir "" svg $ reduce t eqs
	  	          _ -> putStrLn "reduceTerm undefined"
	     svg i = svgTreeSubs $ if even i then green else red
	     g tps rules = init tps ++ iteration n rules (fst $ last tps) []
	     reduce t eqs = if length s > 1 then tail s else s where
	                    s = foldl g [(t,[])] $ Just:map applyEqs eqs++rules

-- iteration n t rules performs at most n successive transformations of t via
-- applications of eval or rules.

iteration :: Int -> Rules String -> Tree String -> [Node] 
					        -> [(Tree String,[Node])]
iteration 0 _ t _      = [(showBin t,[])] 
iteration n rules t ps = if t /= u then f u qs 
                         else if t /= v then f v rs else [(showBin t,[])]
		         where (u,qs) = redStep eval [] t
		               (v,rs) = redStep rules [] t
		               f u qs = (t,ps):(t,qs):iteration (n-1) rules u qs

-- redStep rules p t searches for the outermost redices of the subtree of t at 
-- position p and performs the corresponding reduction steps in parallel.
-- For executing lambda-applications as soon as possible (see eval below), 
-- redices within lambda-abstractions are not taken into account.

redStep :: Rules String -> Node -> Tree String -> (Tree String,[Node])
redStep rules p t = case rules t of 
		    Just u -> (u,[p])
		    _ -> case t of 
		         F a ts | a /= "&#955;" -> (F a us,concat pss) 
		                    where tps = zipWithSucs (redStep rules) p ts 
		                          (us,pss) = unzip tps
		         _ -> (t,[])

applyEqs :: Equations String -> Rules String
applyEqs eqs t = msum $ map f eqs  
	 where f (left,right) = msum [do sub <- match t left
		   			 Just $ right >>> sub,
		   		      do F f [t] <- Just t
		   		         sub <- match (F f []) left
		   		         Just $ F "$" [right >>> sub,t],
		   		      do F f ts <- Just t
		   		         sub <- match (F f []) left
		   		         Just $ F "$" [right >>> sub,F "()" ts],
		   		      do F "$" [t,u] <- Just t
		   		         sub <- match t left
		   			 Just $ F "$" [right >>> sub,u]]

match :: Tree String -> Tree String -> Maybe (String -> Tree String)
match t (V a) = Just $ update V a t
match (F a ts) (F b us) 
               | a == b && length ts == length us = g ts us
               | isJust n  && b == "suc" = do n <- n; g [F (show $ n-1) []] us
               | a == "[]" && b == ":" || a `elem` infixes && a == b       
               		 = do t:ts <- Just ts; [_,_] <- Just us; g [t,F a ts] us 
	         where n = parse nat a
	               g (t:ts) (u:us) = do sub1 <- match t u; sub2 <- g ts us
	    		                    guard $ map sub1 zs == map sub2 zs
	    		                    Just $ \x -> if x `elem` xs 
				         	         then sub1 x else sub2 x 
		                         where xs = vars u
		                               zs = meet xs $ unionMap vars us
                       g _ _ = Just V
match (E a t) (E b u) = do guard $ a == b; match t u
match _ _             = Nothing

mkTrue  = F "true" []
mkFalse = F "false" []

mkBool :: Bool -> Maybe (Tree String) 
mkBool True = Just $ mkTrue
mkBool _    = Just $ mkFalse

mkInt :: Int -> Tree String
mkInt i = F (show i) []

mkReal :: Float -> Tree String
mkReal r = F (show r) []

jmkInt :: Int -> Maybe (Tree String)
jmkInt = Just . mkInt

mkList :: [Tree String] -> Maybe (Tree String)
mkList = Just . F "[]" 

getField :: String -> [Tree String] -> Maybe (Tree String)
getField d (t:ts) = case t of E d' t | d == d' -> Just t
			      _ -> getField d ts
getField d _      = Nothing
		
eval,flatten,impl :: Rules String

eval (F "\\/" ts)   | all isJust bs       = mkBool $ or $ map fromJust bs
                                            where bs = map parseBool ts
eval (F "/\\" ts)   | all isJust bs    	  = mkBool $ and $ map fromJust bs
                                            where bs = map parseBool ts
eval (F "not" [t])  | isJust b		  = mkBool $ not $ fromJust b
					    where b = parseBool t
eval (F "==" [t,u]) | t == u              = mkBool True
eval (F op [t,u])   | isJust rel && isJust i && isJust k
				          = mkBool $ fromJust rel (fromJust i)
				     		   		  (fromJust k) 
				            where rel = parse relation op
				                  i = parseInt t; k = parseInt u
eval (F "suc" [t]) | isJust i             = jmkInt $ fromJust i+1 
					    where i = parseInt t
eval (F "-" [t,u]) | isJust i && isJust k = jmkInt $ fromJust i-fromJust k
			                    where i = parseInt t; k = parseInt u
eval (F "^" [t,u]) | isJust i && isJust k = jmkInt $ fromJust i^fromJust k
			                    where i = parseInt t; k = parseInt u
eval (F "+" ts@(_:_:_)) | all isJust is   = jmkInt $ sum $ map fromJust is
				            where is = map parseInt ts
eval (F "*" ts@(_:_:_)) | all isJust is   = jmkInt $ product $ map fromJust is
				            where is = map parseInt ts
eval (F ":" [t,F "[]" ts])                = mkList $ t:ts
eval (F "id" [t])                         = Just t
eval (F "ite" [t,u,v]) | isJust b         = Just $ if fromJust b then u else v
					    where b = parseBool t
eval (F "range" [t,u]) | isJust i && isJust k  
                                          = mkList $ map mkInt [fromJust i..
                                          			fromJust k]
			                    where i = parseInt t; k = parseInt u
eval (F "$" [F "lsec" [t,F op []],u])  	  = Just $ F op [t,u]
eval (F "$" [F "rsec" [F op [],t],u])     = Just $ F op [u,t]
eval (F "$" [F "." [t,u],v])              = Just $ F "$" [t,F "$" [u,v]]
eval (F "$" [F "&#955;" [F "->" [p,t]],u]) 		  -- lambda-applications
 			     | isJust sub = Just $ t>>>fromJust sub
					    where sub = match u p
eval (F "$" [F "&#955;" ts@(_:_:_),u])      
                          | not $ null us = Just $ F "CASE" $ map fromJust us
	         where us = filter isJust $ map f ts
                       f (F "->" ts) = msum [do [p,t] <- Just ts
      				                sub <- match u p
    				                Just $ F "->" [mkTrue,t>>>sub],
			   	             do [p,b,t] <- Just ts
			   	                sub <- match u p
			   	                Just $ F "->" [b>>>sub,t>>>sub]]
eval (F "$" [F f [],F "()" ts])              = Just $ F f ts
eval (F "$" [t,F "()" [u]])                  = Just $ F "$" [t,u]
eval (F "$" [F f [],t]) 		     = Just $ F f [t]
eval (F "CASE" (F "->" [F "true" [],t]:_))   = Just t
eval (F "CASE" (F "->" [F "false" [],_]:ts)) = Just $ F "CASE" ts
eval (F "CASE" []) 			     = Just $ F "undefined" []
eval (F d [F _ ts@(E _ _:_)])    	     = getField d ts
eval (F "&" ts@(_:_:_:_))                    = Just $ F "$" [last ts,
							     F "&" $ init ts]
eval (F "&" [t,u]) 		  	     = Just $ F "$" [u,t]
eval _					     = Nothing
	       
flatten t = do F op ts <- Just t
               guard $ op `elem` words "\\/ /\\ + *" 
               let u = norm op ts
               guard $ u /= t
	       Just u
	       
showBin t@(F "Bjoin" [_,_,_]) = f t where
		                f (F "Bjoin" [a,l,r]) = F (root a) [f l,f r]
		                f (F "Empty" _)       = F "o" []
                                f _                   = F ".." []
showBin t = t
		
impl = \case F "\\/" ts -> if b then mkBool True 
                 	   else Just $ F "==>" [norm "/\\" prem,norm "or" conc]
                 	   where (prem,conc,b) = split ts
	     t@(F "/\\" ts) | t /= u -> Just u where u = norm "/\\" ts
	     _ -> Nothing
       where split (t:ts) = if root t == "not" 
		            then if u `elem` vs then empty else (u:us,vs,b) 
			    else if t `elem` us then empty else (us,t:vs,b) 
		            where empty = ([],[],True)
		          	  (us,vs,b) = split ts
		          	  u = head $ subtrees t
             split _      = ([],[],False)

zero "*"   = mkTrue
zero "/\\" = mkTrue
zero _     = mkFalse
	       
norm :: String -> [Tree String] -> Tree String
norm op []  = zero op
norm op [t] = t
norm op ts  = case args (F op ts) of [] -> zero op; [t] -> t; ts -> F op ts
         where args (F x ts) | x == op = unionMap args $ filter (/= zero op) ts
               args t	               = if t == zero op then [] else [t]

distribute :: String -> String -> Rules String
distribute a b (F c ts) | a == c && isJust ti 
		 = Just $ F b $ map (F a . updList ts i) us 
		   where ti = searchGet ((== b) . root) ts
		         (F _ us,i) = fromJust ti
distribute _ _ _ = Nothing

-- Examples

attr2   = reduce "attr2" "z s" [] 
dist1   = reduce "dist1" "" [distribute "*" "+",flatten] 		   
dist2   = reduce "dist2" "x y z" [flatten] 				   
eval1   = reduce "eval1" "x y" []
eval2   = reduce "eval2" "x y z z'" [flatten] 
fact    = reduce "fact" "x" [] 
factI   = reduce "factI" "x y" [] 
fiter   = reduce "fiter" "n F" [] 
fibs    = reduce "fibs" "x y n s s' F" [] 
fold    = reduce "fold" "x y s F" [] 
gentzen = reduce "gentzen" "x y z" [distribute "\\/" "/\\",flatten,impl]
iterat  = reduce "iterat" "x n s F" [] 
lamb    = reduce "lamb" "x y z F" [] 
mapzip  = reduce "mapzip" "x y s s' F" [] 

unfold1 = reduceTerm 50 "unfoldList" "x s val" []
unfold2 = reduceTerm 50 "unfoldBin" "a b l r t u n val" []

-- XCTL = XPath and CTL: subsets of and binary relations between tree nodes

data XCTL nodeSet nodeRel = 
     XCTL {true_,false_       :: nodeSet,
     	   atom	              :: (String -> Bool) -> nodeSet,
           not_,ex,ax,af,eg   :: nodeSet -> nodeSet,
     	   and_,or_           :: nodeSet -> nodeSet -> nodeSet,
     	   exists,forall,div_ :: nodeRel -> nodeSet -> nodeSet,
     	   self,child,parent,next,prev,equal,equiv 
     	   		      :: nodeRel,
     	   closure,inv        :: nodeRel -> nodeRel,
     	   par,seq_	      :: nodeRel -> nodeRel -> nodeRel,
     	   restrict	      :: nodeRel -> nodeSet -> nodeRel}

xalg :: Tree String -> XCTL [Node] (Node -> [Node])
xalg t = XCTL {true_     = nodes t,
               false_    = [],
               atom      = \pred -> filter (pred . label t) $ nodes t,
               not_      = diff $ nodes t,
               and_      = meet,
               or_       = union,
               exists    = exists,
               forall    = forall,
               div_      = div_,
               ex        = ex',
               ax        = ax',
               af 	 = \s -> fixpt subset (stepAF s) [],
	       eg        = \s -> fixpt supset (stepEG s) $ nodes t,
               self      = self,
               child     = child,
               parent    = parent,
               next      = next,
               prev      = prev,
               equal     = rel2fun $ fixpt supset stepEqual nodePairs,
               equiv     = rel2fun $ fixpt supset stepEquiv nodePairs,
               closure   = \r -> fixpt le ((r#) . (r>=>)) r,
               inv 	 = \r -> rel2fun $ fun2invrel r $ nodes t,
               par       = (#), 
               seq_      = (>=>),
               restrict  = \r -> \s -> \w -> r w `meet` s}
         where (#)           = liftM2 union
               self          = return
               child v       = [w | w <- nodes t, not $ null w, init w == v]
               parent []     = []
               parent w      = [init w]
               next []       = [] 
               next w        = if v `elem` nodes t then [v] else []
               		       where v = init w++[last w+1]
               prev []       = [] 
               prev w        = if v `elem` nodes t then [v] else []
               		       where v = init w++[last w-1]
	       exists r s    = [w | w <- nodes t, r w `shares` s]
	       forall r s    = [w | w <- nodes t, r w `subset` s]
               div_ r s      = [w | w <- nodes t, s `subset` r w]
               nodePairs     = [(v,w) | v <- s, w <- s, v < w] where s = nodes t
               stepEqual rel = filter f nodePairs where
               	               f (v,w) = label t v == label t w &&
	  		                 length (child v) == length (child w) &&
	                                 zip (child v) (child w) `subset` rel
               stepEquiv rel = filter f nodePairs where
               	               f (v,w) = label t v == label t w &&
	  		                 liftR rel (child v) (child w)
	       ex' 	     = exists child
	       ax'           = forall child
	       stepAF s s'   = union s $ meet (ax' s') $ ex' $ nodes t
	       stepEG s s'   = meet s $ union (ex' s') $ ax' []
	       le r r'       = all (liftM2 subset r r') $ nodes t

-- derived operators	  

descendant alg = closure alg $ child alg
ancestor alg   = closure alg $ parent alg
folSib alg     = closure alg $ next alg
preSib alg     = closure alg $ prev alg
following alg  = seq_ alg (psa alg) $ seq_ alg (folSib alg) $ psd alg
preceding alg  = seq_ alg (psa alg) $ seq_ alg (preSib alg) $ psd alg
ef alg         = exists alg $ psd alg  
ag alg         = forall alg $ psd alg 

psa alg = par alg (self alg) $ ancestor alg
psd alg = par alg (self alg) $ descendant alg
	       			       
foldSet :: XCTL nodeSet nodeRel -> Tree String -> Maybe nodeSet
foldSet alg = f where 
        f = \case F "true" []       -> Just $ true_ alg
                  F "false" []      -> Just $ false_ alg   
                  F a []            -> Just $ atom alg (== a)   
                  F "rsec" [F op [],F a []] | isJust i && isJust rel      
                      	            -> Just $ atom alg p where
                     		       i = parse int a                     			  
                     		       rel = parse relation op
                     		       p b = isJust k && fromJust rel 
                     		       		       (fromJust k) (fromJust i)
                       			     where k = parse int b
                  F "/\\" ts@(_:_)  -> fmap (foldl1 $ and_ alg) $ mapM f ts
                  F "\\/" ts@(_:_)  -> fmap (foldl1 $ or_ alg) $ mapM f ts
                  F "exists" [t,u]  -> liftM2 (exists alg) (foldRel alg t) $ f u
                  F "forall" [t,u]  -> liftM2 (forall alg) (foldRel alg t) $ f u
                  F "div" [t,u]     -> liftM2 (div_ alg) (foldRel alg t) $ f u
                  F "not" [t]	    -> fmap (not_ alg) $ f t
                  F "EX" [t]	    -> fmap (ex alg) $ f t
                  F "AX" [t]        -> fmap (ax alg) $ f t
                  F "EF" [t]        -> fmap (ef alg) $ f t
                  F "AF" [t]        -> fmap (af alg) $ f t
                  F "EG" [t]        -> fmap (eg alg) $ f t
                  F "AG" [t]        -> fmap (ag alg) $ f t
                  _ 		    -> Nothing

foldRel :: XCTL nodeSet nodeRel -> Tree String -> Maybe nodeRel
foldRel alg = f where 
         f = \case F "+" ts@(_:_) -> fmap (foldl1 $ par alg) $ mapM f ts
                   F "/" ts@(_:_) -> fmap (foldl1 $ seq_ alg) $ mapM f ts
                   F "clos" [t]   -> fmap (closure alg) $ f t
                   F "inv" [t]    -> fmap (inv alg) $ f t
                   F ">>" [t,u]   -> liftM2 (restrict alg) (f t) $ foldSet alg u
                   F x []         -> case x of 
                                          "self"       -> Just $ self alg
                                          "child"      -> Just $ child alg
                                          "parent"     -> Just $ parent alg
                                          "next"       -> Just $ next alg
                                          "prev"       -> Just $ prev alg
                                          "descendant" -> Just $ descendant alg
                                          "ancestor"   -> Just $ ancestor alg
                                          "folSib"     -> Just $ folSib alg
                                          "preSib"     -> Just $ preSib alg
                                          "following"  -> Just $ following alg
                                          "preceding"  -> Just $ preceding alg
                                          "equal"      -> Just $ equal alg
                                          "equiv"      -> Just $ equiv alg
                                          _ -> Nothing
                   _              -> Nothing
                      
-- Examples 
       
rel1 = "(child >> (a /\\ exists(descendant,d))) / child / (descendant >> c)"

showRel = mkTree "rel1" rel1  

testrel1 = drawNodeRel "baum9" rel1 	 	
testrel2 = drawNodeRel "baum9" "preceding"   	 	
testrel3 = drawNodeRel "baum9" "following" 	 	
testrel4 = drawNodeRel "baum9" "equal" 
testrel5 = drawNodeRel "baum9" "equiv" 	 	

testset1 = drawNodeSet "baum4" "EX(true)\\/88" 
testset2 = drawNodeSet "tree3" "EF(63)"      
testset3 = drawNodeSet "tree3" "AF(63)" 
testset4 = drawNodeSet "baum4" "EG((<=22)\\/66)" 
testset5 = drawNodeSet "baum4" "AG((<=22)\\/66)" 

-- SQL (see https://de.wikipedia.org/wiki/SQL,	
--	    https://de.wikipedia.org/wiki/Relationale_Algebra,
--          https://en.wikipedia.org/wiki/Relational_algebra)

data SQL table row a = 
     SQL {project 	  :: [String] -> table -> table,
     	  select  	  :: (row -> Bool) -> table -> table,
     	  prod,div',njoin :: table -> table -> table,
     	  ejoin           :: String -> String -> table -> table -> table,
     	  getAttrs	  :: table -> [String],
     	  getTable	  :: String -> IO table,
     	  getfromDB	  :: String -> String -> IO table,
     	  pictTerm        :: table -> String -> String,
     	  putHere         :: table -> IO table,
     	  putPict	  :: table -> String -> IO table,
     	  emptyTab        :: table,
     	  mkTab 	  :: [row] -> table,
     	  mkRows	  :: table -> [row],
     	  getVal	  :: String -> row -> a,
     	  readVal         :: Parser a,
     	  errVal          :: a,
     	  addVal          :: a -> a -> a}

type Table a = Map String [a]
type Row a   = Map String a 

showTab :: Show a => Table a -> String
showTab tab = init ("tab{" ++ drop 5 (concatMap g $ keys tab)) ++ "}"
              where g x = "\n    " ++ x ++ " = " ++ show (tab!x) ++ ","
              
-- a sum type for attribute values

data Val = N String | I Int | R Float deriving (Eq,Ord)

attrVal :: Parser Val
attrVal = msum [do name <- identifier; return $ N name,
		do r <- real; return $ R r,
		do i <- int; return $ I i]

instance Show Val where show (N name) = name
			show (I i)    = show i
			show (R r)    = show r
     	     
sql :: SQL (Table Val) (Row Val) Val
sql = SQL {project  = project,  select = select, prod = prod, div' = div',
	   njoin    = njoin,    ejoin  = ejoin,  getAttrs = keys,
	   getTable = getTable, getfromDB = getfromDB, pictTerm = pictTerm,
 	   putHere  = putHere,  putPict   = putPict,   emptyTab = emptyTab, 
 	   mkTab    = mkTab,    mkRows    = mkRows,    getVal   = flip (!), 
 	   readVal  = readVal,  errVal    = errVal,    addVal   = addVal}
      where project xs tab   = filterWithKey f tab where f x _ = x `elem` xs
	    select pred tab  = mkTab $ filter pred $ mkRows tab
	    ptab p tab1 tab2 = mkTab [DMS.union row1 row2 | 
	         			        row1 <- mkRows tab1,
	         			        row2 <- mkRows tab2,p row1 row2]
            prod tab1 tab2  = ptab (\_ _ -> True) tab1 $ DMS.union tab3 tab4
		              where xs = keys tab1; ys = meet xs $ keys tab2
		                    zs = map ('!':) ys               -- renaming
		                    tab3 = foldl (flip delete) tab2 xs
		                    tab4 = fromList $ zip zs $ map (tab2!) ys
	    div' tab1 tab2  = mkTab [row | row <- rows, all (p row) rows2] 
	                      where xs = keys tab1 `diff` keys tab2
	                            rows = mkSet $ mkRows $ project xs tab1
	                            rows1 = mkRows tab1
	                            rows2 = mkRows tab2
	                            p row row2 = DMS.union row row2 `elem` rows1
	    njoin = ptab p where m row1 row2 = keys row1 `meet` keys row2
	                         p row1 row2 = all f $ m row1 row2
	                                       where f x = row1!x == row2!x
	    ejoin x y = ptab p where p row1 row2 = notMember x row1 || 
	         				   notMember y row2 ||
	         				   row1!x == row2!y
	    getTable file = readFileContinue emptyTab file act 
		                where act str = case parse (term mzero) str of 
		     		                Just (F "tab" ts) 
		     		                  -> case trees2cols ts of 
		    		      	                  Just tab -> return tab
		    		      	                  _ -> return emptyTab
		    		                _ -> return emptyTab
	    getfromDB tab file = readFileContinue emptyTab file act 
		                where act db = case parse (term mzero) db of 
		     		               Just (F "db" ts) 
		     		                 -> case lookupTab tab ts of 
		    		      	                 Just tab -> return tab
		    		      	                 _ -> return emptyTab
		    		               _ -> return emptyTab

            -- lookupTab ts takes the table t with name tab from a list of trees
            -- representing tables and transforms t into an object of type
            -- Table Val.

            lookupTab :: String -> [Tree String] -> Maybe (Table Val)
            lookupTab tab ts = do E tab' (F "tab" ts):us <- Just ts
                                  if tab == tab' then trees2cols ts
                                  		 else lookupTab tab us

            trees2cols :: [Tree String] -> Maybe (Table Val)
            trees2cols ts = do pairs <- mapM tree2col ts; Just $ fromList pairs

            tree2col :: Tree String -> Maybe (String,[Val])
            tree2col t = do E x (F "[]" ts@(_:_)) <- Just t
            		    as <- mapM tree2val ts; Just (x,as)

            tree2val :: Tree String -> Maybe Val
            tree2val t = do F x [] <- Just t; parse readVal x
                 
            -- pictTerm tab name constructs a command term t from tab. t gets
            -- Expander2 to paint tab as a matrix.
		         
            pictTerm tab name = "shelf(1)[text$" ++ name ++ ",white$circ$4,mat" 
            			++(f $ show $ concat $ zipWith g inds rows)++"]"
		                where rows = mkRows tab
		                      inds = [1..length rows]
		                      f ('"':str) = f str
		                      f (x:str)   = x:f str
		                      f _ 	 = ""
		                      g i row = [(i,x,row!x) | x <- keys row]
		       
 	    putHere tab      = do putStrLn $ showTab tab; return tab
	    putPict tab name = do writeFile ("../ExpanderLib/"++name) 
		 		            $ pictTerm tab name
                 		  return tab
		          
            emptyTab = DMS.empty
                 
            mkTab (row:rows) = foldl f (mkTab rows) $ keys row
	               where f tab x = case DMS.lookup x tab of
	                                    Just s -> DMS.insert x (row!x:s) tab
	                                    _      -> DMS.insert x [row!x] tab
            mkTab _ = emptyTab 
            
            mkRows :: Table Val -> [Row Val]
            mkRows tab = if DMS.null tab then [] else map g [0..max-1]
                         where xs = keys tab
                               g i = fromList $ map (\x -> (x,h x i)) xs
                               h x i = if i < length tx then tx!!i else errVal
                                       where tx = tab!x
		               max = maximum $ map (length . (tab!)) xs 

            readVal = msum [do name <- identifier; return $ N name,
                 	    do r <- real; return $ R r,
                 	    do i <- int; return $ I i]
                 	         
            errVal = N "no value"	 
                 
            addVal (I i) (I k) = I $ i+k
            addVal (I i) (R r) = R $ fromIntegral i+r
            addVal (R r) (I k) = R $ r+fromIntegral k
            addVal _ _ 	       = errVal
                 		       
foldTable :: Ord a => SQL table row a -> Tree String -> MaybeT IO table
foldTable alg = f where
       f = \case F "proj" [F "[]" ts,t] | all isLeaf ts
                	             -> fmap (project alg $ map root ts) $ f t
                 F "proj" [t,u] | isLeaf t 
                                     -> fmap (project alg [root t]) $ f u
                 F "sel" [p,t]       -> do tab <- f t
                	                   let xs = getAttrs alg tab
                	  	           pred <- liftT $ foldPred alg xs p 
                	  	           return $ select alg pred tab
                 F "prod" [t,u]      -> liftM2 (prod alg) (f t) $ f u
                 F "div" [t,u]       -> liftM2 (div' alg) (f t) $ f u
                 F "tjoin" [p,t,u]   -> f $ F "select" [p,F "prod" [t,u]]
                 F "njoin" [t,u]     -> liftM2 (njoin alg) (f t) $ f u
                 F "ejoin" [x,y,t,u] -> liftM2 (ejoin alg (root x) $ root y)
                    		               (f t) $ f u
                 F "here" [t]        -> do tab <- f t; lift $ putHere alg tab
                 F "pict" [t,u] | isLeaf u
                	  	     -> do tab <- f t
                	  	           lift $ putPict alg tab $ root u
                 F "fromDB" [t,u] | isLeaf t && isLeaf u
                	  	     -> lift $ getfromDB alg (root t) $ root u
                 F "$" [F g [],t]    -> f $ F g [t]
                 t | isLeaf t        -> lift $ getTable alg $ root t
                 _		     -> liftT Nothing
                 
foldPred :: Ord a => SQL table row a -> [String] -> Tree String 
					         -> Maybe (row -> Bool)
foldPred alg xs = f 
          where f = \case F "\\/" ts@(_:_) -> fmap (foldl1 (|||)) $ mapM f ts
                          F "/\\" ts@(_:_) -> fmap (foldl1 (&&&)) $ mapM f ts
                          F "not" [t]      -> fmap (fmap not) $ f t
 		   	  F op [t,u] | isJust rel 
 		   	 		   -> liftM2 (liftM2 rel') (ff t) $ ff u
	          	     		      where ff = foldFun alg xs
	          	     		            rel = parse relation op
	          	     		            rel' = fromJust rel
	          	  _ 		   -> Nothing
	          	     
foldFun :: SQL table row a -> [String] -> Tree String -> Maybe (row -> a)
foldFun alg xs = f 
     where f = \case F x [] -> Just $ if x `elem` xs then getVal alg x
                                      else const $ case parse (readVal alg) x of
                                                       Just a -> a
                                                       _ -> errVal alg
             	     F "+" ts@(_:_:_) -> fmap (foldl1 $ liftM2 $ addVal alg) 
             	  			      $ mapM f ts
 	             _ -> Nothing
	          	    
-- compTable query parses the term query and folds the resulting table.

compTable :: String -> IO ()
compTable query = case parse (term mzero) query of 
		       Just tree -> do tab <- runMT $ foldTable sql tree
		                       if isJust (tab :: Maybe (Table Val)) 
		                       then return () else putStrLn "NO TABLE"
	    	       _         -> putStrLn "NO PARSE"

-- dbPict file parses the database term db stored in file, transforms db into a
-- list of widget terms representing the tables of db according to pictTerm and
-- saves them to ExpanderLib/fileT. Enter loadT$fileT into the canvas of
-- Expander2's solver window and press the paint button for displaying the 
-- tables of db graphically.

dbPict :: String -> IO ()
dbPict file = readFileContinue () file act where
      act db = case parse (term mzero) db of 
	       Just (F "db" ts) | all (fst . isTable) ts
	         -> do tabs <- mapM (f . isTable) ts
		       let terms = zipWith (pictTerm sql) tabs 
		      			   $ map (snd . isTable) ts
		       writeFile ("../ExpanderLib/" ++ file ++ "T")
		                 $ '[':drop 2 (concatMap (",\n"++) terms) ++ "]"
		    where isTable = \case E x _ -> (True,x)
		            	          _ -> (False,"")
		          f (_,tab) = getfromDB sql tab file :: IO (Table Val)
               _ -> putStrLn "NO DATABASE"

-- Examples 

db1    = dbPict "database"
drawDB = drawTermC "database"
drawT  = drawTerm "table1"
drawM  = compTable "pict(table1,table1)"
		         
table1 = compTable "here$table1" 	-- tab{name = [Hilde,Inge,Robin],
    				     	--     age = [81,85,72],
    				     	--     children = [1,3,2],
    				     	--     temp = [5.66,9.8,66.09]}

table2 = compTable "here$fromDB(table2,database)"	
					-- tab{name = [Robin,Inge,Markus],
    					--     wife = [Hanna,Isolde,Hilde],
    					--     age = [72,85,72]} 	

table3 = compTable "here$sel(age<78+2\\/name==Hilde,table1)"
					-- tab{name = [Hilde,Robin],
					--     age = [81,72],
					--     children = [1,2],
					--     temp = [5.66,66.09]

table4 = compTable "here$proj([name,age],table1)"	
					-- tab{name = [Hilde,Inge,Robin],
					--     age = [81,85,72]}	

table5 = compTable "here$proj([name,children],table1)"	
					-- tab{name = [Hilde,Inge,Robin],
					--     children = [1,3,2]}	

table6 = compTable "here$njoin(table1,fromDB(table2,database))" 
					-- tab{name = [Inge,Robin], 
					--     age = [85,72],	    
					--     children = [3,2],
					--     wife = [Isolde,Hanna],
					--     temp = [9.8,66.09]}	
   
table7 = compTable "here$fromDB(table7,database)"	
					-- tab{A = [1,4,7],
    					--     B = [2,5,8],
    					--     C = [3,6,9],
    					--     D = [4,7,0]}		
   
table8 = compTable "here$fromDB(table8,database)"	
					-- tab{A = [1,7],
    					--     F = [2,8],
    					--     G = [3,9]}

table9 = compTable $ "here$njoin(fromDB(table7,database)," ++
		     		"fromDB(table8,database))"
			                -- tab{A = [1,7],	
         				--     B = [2,8],
         				--     C = [3,9],
         				--     D = [4,0],
         				--     F = [2,8],
         				--     G = [3,9]}
    					
table10 = compTable $ "here$fromDB(table10,database)"	
					-- tab{E = [1,7],
    					--     F = [2,8],
    					--     G = [3,9]}

table11 = compTable $ "here$ejoin(C,G,fromDB(table7,database)," ++
				 "fromDB(table10,database))"		
    				        -- tab{A = [1,7],
         				--     B = [2,8],
         				--     C = [3,9],
         				--     D = [4,0],
         				--     E = [1,7],
         				--     F = [2,8],
         				--     G = [3,9]}

table12 = compTable $ "here$div(fromDB(completed,database)," ++
		               "fromDB(dbProject,database))"		
		             		-- tab{Student = [Fred,Sarah]}	
			       
table13 = compTable $ "here$div(fromDB(table11,database)," ++
			       "fromDB(table12,database))"   		
			     	        -- tab{Vater = [Moritz],
          				--     Mutter = [Melanie]}	

table14 = compTable $ "here$div(fromDB(table14,database)," ++
			       "fromDB(table15,database))"
			                -- tab{A = [1],
          		  	       	--     B = [2]}
	           
table15 = compTable $ "here$prod(proj([A,B,C],fromDB(table7,database))," ++
	           	        "fromDB(table13,database))" 	
	           			-- tab{A = [1,1,4,4,7,7],
	                    		--     B = [2,2,5,5,8,8],
	                    		--     C = [3,3,6,6,9,9],
	                    		--     D = [3,6,3,6,3,6],
	                    		--     E = [1,2,1,2,1,2]}
		
query1 = "here$tjoin(B<D,proj([A,B,C],fromDB(table7,database))," ++
	            "fromDB(table13,database))" 

table16 = compTable query1		-- tab{A = [1,1,4],
					--     B = [2,2,5],
					--     C = [3,3,6],
					--     D = [3,6,6],
					--     E = [1,2,2]}

drawQ1 = mkTree "query1" query1

-- TREES WITH NODE POSITIONS

type PTree = Tree (String,Point)

nodeSize :: String -> Point
nodeSize a = (w,h) where (_,_,w,h) = textblock p0 a
	  
nodeWidth :: String -> Float
nodeWidth = fst . nodeSize	

-- mkPTree{'} p (hor,ver) t adds positions in the plane to the nodes of t such 
-- that p is the leftmost-uppermost corner of the smallest rectangle enclosing 
-- t. hor and ver are the horizontal resp. vertical spaces between adjacent 
-- nodes. 

mkPTree :: Point -> Point -> Tree String -> PTree
mkPTree (hor,ver) p = alignLeaves hor . f p where
      f (x,y) (V a)    = V (a,(x+nodeWidth a,y))
      f (x,y) (F a []) = F (a,(x+nodeWidth a,y)) []
      f (x,y) (F a ts) = if diff <= 0 then qt else translate diff qt
                         where diff = nodeWidth a-fst (treeSize pt)/2
			       hdiff = height a+maximum (map (height . root) ts)
			       height a = h+h where (_,_,_,h) = textblock p0 a
			       pt:pts = map (f (x,y+ver+hdiff/2)) ts
			       qts = transTrees hor pt pts
			       qt = F (a,((g (head qts)+g (last qts))/2,y)) qts
			       g = fst . snd . root
      f p (E a t)      = E ap pt where F ap [pt] = f p (F a [t])

-- translate d t moves t by d units to the right.

translate :: Float -> PTree -> PTree
translate d = mapTree f where f (a,(x,y)) = (a,(x+d,y))

-- transTrees hor pt pts orders the trees of pt:pts with a horizontal space of 
-- hor units between adjacent trees. transTrees takes into account different 
-- heights of adjacent trees by shifting them to the left or to the right such 
-- that nodes on low levels of a tree may occur below a neighbour with fewer 
-- levels.

transTrees :: Float -> PTree -> [PTree] -> [PTree]
transTrees hor pt = f [pt]
         where f us (t:ts) = if d < 0 then f (map (translate $ -d) us++[t]) ts
                             else f (us++[translate d t]) $ map (translate d) ts
                          -- f (us++[if d < 0 then t else translate d t]) ts 
                             where d = maximum (map h us)+hor
	      	                   h u = f (+) maximum u-f (-) minimum t
			                 where f = g $ min (height t) $ height u
               f us _      = us
               g _ op _ (V (a,(x,_)))    = h op x a
               g 1 op m (F (a,(x,_)) _)  = h op x a
               g n op m (F (a,(x,_)) ts) = m $ h op x a:map (g (n-1) op m) ts
               g n op m (E (a,(x,_)) t)  = m [h op x a,g (n-1) op m t]
	       h op x = op x . nodeWidth

-- textblock interprets '%' as a line break and thus admits multiline node 
-- labels.	 

textblock :: Point -> String -> (Path,[String],Float,Float)
textblock (x,y) a = (map f [0..upb],ls,maximum $ map lgW ls,h)
                    where f i = (x,y-h+20*float i)
                    	  ls = lines a [] ""
                    	  h = 10*float upb
		    	  upb = length ls-1
		    	  lines (x:'%':[]) ls b = ls++[b++[x,'%']]
		    	  lines (x:'%':a) ls b  = lines a (ls++[b++[x]]) ""
		    	  lines (x:a) ls b      = lines a ls $ b++[x]
			  lines _ ls b          = ls++[b]
			  
treeLeft :: PTree -> Float
treeLeft = minimum . bounds
           where bounds t = f (root t) $ map bounds $ subtrees t
                 f :: (String,Point) -> [[Float]] -> [Float]
	         f (a,(x,y)) fss = x+nodeWidth a:concat fss
			  
treeSize :: PTree -> Point
treeSize = max2 . foldTree bounds
           where bounds :: (String,Point) -> [Path] -> Path
	         bounds (a,p) pss = add2 p (nodeSize a):concat pss

-- alignLeaves hor t replaces the leaves of t such that all horizontal gaps 
-- between neighbours become equal.

alignLeaves :: Float -> PTree -> PTree
alignLeaves hor (F a [t]) = F a [alignLeaves hor t]
alignLeaves hor (F a ts)  = F a $ equalGaps hor $ map (alignLeaves hor) ts 
alignLeaves hor (E a t)   = E a $ alignLeaves hor t
alignLeaves _ t           = t

equalGaps :: Float -> [PTree] -> [PTree]
equalGaps hor ts = if length ts > 2 then us++vs else ts
                   where (us,vs) = foldl f ([],[head ts]) $ tail ts
                         f (us,vs) v = if isLeaf v then (us,vs++[v])
	         	               else (us++transLeaves hor vs v,[v])
	
transLeaves :: Float -> [PTree] -> PTree -> [PTree]
transLeaves hor ts t = loop hor
             where loop hor = if x1+w1+hor >= x2-w2 then us else loop $ hor+1 
		        where us = transTrees hor (head ts) $ tail ts
		              [x1,x2] = map (fst . snd . root) [last us,t]
		              [w1,w2] = map (nodeWidth . fst . root) [last us,t]

-- svgTree{C} spread t computes SVG code for the tree of type PTree each of
-- whose node labels includes the position where the node is to be placed.

svgTree,svgTreeC :: Svg (Tree String)

svgTree spread t = (f pt,round2 $ treeSize pt)
	     where (_,h) = nodeSize $ root t
	           pt = mkPTree spread (50,20+h) t
		   f :: PTree -> String
		   f (V (a,p))    = concat $ nodes a p
	           f (F (a,p) ts) = concat $ map (g p) ts ++ nodes a p
	           f (E (a,p) t)  = concat $ g p t:nodesE a p
	           g p t = line red p (snd $ root t)++f t
	           nodes a p  = zipWith (drawBox rectangle $ light yellow) ps ls
			        where (ps,ls,_,_) = textblock p a
	           nodesE a p = zipWith (drawBox rectangle white) ps ls
			        where (ps,ls,_,_) = textblock p a

svgTreeC spread t = (f red pt,round2 $ treeSize pt)
  where pt = mkPTree spread (50,20) t
        f :: RGB -> PTree -> String
        f col (V (a,p))    = concat $ nodes (light col) ps ls
	  		     where (ps,ls,_,_) = textblock p a
        f col (F (a,p) ts) = concat $ map (g p next) ts++nodes (light col) ps ls
                             where (ps,ls,_,_) = textblock p a
                                   next = nextColor (height t) col
 	f col (E (a,p) t)  = concat $ g p col t:nodes white ps ls
                             where (ps,ls,_,_) = textblock p a
	g p col t = line col p (snd $ root t)++f col t
	nodes = zipWith . drawBox ellipse

svgTreeP :: Int -> Svg (Tree String,Node)
svgTreeP i spread (t,node) = (f pt [],round2 $ treeSize pt)
      where pt = mkPTree spread (20,20) t
	    f :: PTree -> [Int] -> String
	    f (V (a,p)) pos    = concat $ nodes a p pos
	    f (F (a,p) ts) pos = concat $ zipWithSucs g pos ts++nodes a p pos
                              where g pos t = line red p (snd $ root t)++f t pos
            nodes a p pos = zipWith (drawBox rectangle $ lighter col) ps ls
		        where (ps,ls,_,_) = textblock p a
		              col | i == 1       = yellow
			          | node <<= pos = if even i then red else green
			          | True 	 = yellow

svgTreeSubs :: RGB -> Svg (Tree String,[Node])
svgTreeSubs col spread (t,subs) = (f [] pt,round2 $ treeSize pt)
   where (_,h) = nodeSize $ root t
	 pt = mkPTree spread (50,20+h) t
	 f :: Node -> PTree -> String
	 f pos (V (a,p))    = concat $ nodes a p pos
	 f pos (F (a,p) ts) = concat $ zipWithSucs (g p) pos ts ++ nodes a p pos
	 f pos (E (a,p) t)  = concat $ g p (pos++[0]) t : nodesE a p pos
         g p pos t = line red p (snd $ root t)++f pos t
         nodes a p pos  = zipWith (drawBox rectangle col') ps ls 
			  where (ps,ls,_,_) = textblock p a
			        col' = lighter $ if any (<<= pos) subs 
			        		 then col else yellow
         nodesE a p pos = zipWith (drawBox rectangle white) ps ls
			  where (ps,ls,_,_) = textblock p a

svgTreeSet :: Svg (Tree String,[Node])
svgTreeSet spread (t,poss) = (svgTreeSR spread col t [] pt,
			      round2 $ treeSize pt)
                       where pt = mkPTree spread (50,20) t
                             col pos = if pos `elem` poss then green else yellow

svgTreeRel :: Svg (Tree String,Node,[Node])
svgTreeRel spread (t,pos,poss) = (svgTreeSR spread col t [] pt,
				  round2 $ treeSize pt)
	               where pt = mkPTree spread (50,20) t
	                     col pos' = if pos' `elem` poss 
	                    	        then if pos == pos' then blue else green
			                else if pos == pos' then red else yellow
			                            
svgTreeSR :: Point -> (Node -> RGB) -> Tree String -> Node -> PTree -> String
svgTreeSR spread col t = f where 
         f :: Node -> PTree -> String
         f pos (V (a,p))    = concat $ nodes a p pos
	 f pos (F (a,p) ts) = concat $ zipWithSucs (g p) pos ts ++ nodes a p pos
	 f pos (E (a,p) t)  = concat $ g p (pos++[0]) t : nodes a p pos
	 g p pos t = line red p (snd $ root t)++f pos t
         nodes a p pos = zipWith (drawBox ellipse $ lighter $ col pos) ps ls 
        		 where (ps,ls,_,_) = textblock p a

-- GRAPHS

p0 = (0,0)
p1 = (1,1)
p2 = (10,40)

showMode :: Int -> String
showMode mode = if lg < 5 then str++replicate (5-lg) '1' else take 5 str
	        where str = show mode
		      lg = length str  

-- straight p q r checks whether p, q and r form a line

straight :: Point -> Point -> Point -> Bool
straight (x1,y1) (x2,y2) (x3,y3) = x1 == x2 && x2 == x3 || 
                                   x1 /= x2 && x2 /= x3 &&
                                   (y2-y1)/(x2-x1) == (y3-y2)/(x3-x2)

minimize :: Path -> Path
minimize (p:ps@(q:r:s)) = if straight p q r then minimize $ p:r:s
                       			    else p:minimize ps
minimize ps = ps  

center,gravity :: Path -> Point
center ps        = ((x1+x2)/2,(y1+y2)/2) where (x1,y1,x2,y2) = minmax ps
gravity ps@(p:_) = apply2 (/(float $ length qs)) $ foldl1 add2 qs
                   where qs = if p == last ps then tail ps else ps
gravity _        = p0

area :: [(Double,Double)] -> Double
area ps = abs (sum $ zipWith f ps $ tail ps++[head ps])/2
          where f (x1,y1) (x2,y2) = (x1-x2)*(y1+y2)

-- draw complex functions

type ComplexFun = Complex Float -> Complex Float
		
drawGrid :: ComplexFun -> IO ()
drawGrid f = do writeFile "grid" $ show $ mkGrid 50 f
	        readAndDraw p0 "grid" "" svgPlane $ Just . read
								     
mkGrid :: Int -> ComplexFun -> [CPoint]
mkGrid n f = zipWith h grid distances
             where coords = reverse [-n..(-1)]++[0..n]
	           grid = [float2 (x,y) | x <- coords, y <- coords]
                   distances = map g grid
		   g (x,y) = round $ distance (x,y) (a,b)
		                  -- angle 
		             where a:+b = f $ x:+y
		   maxdist = maximum distances
		   h (x,y) d = (hue 1 red maxdist d,x*4,y*4)
	   
svgPlane :: Svg [CPoint]
svgPlane (hor,ver) path = (code,size)
                   where (minx,miny,maxx,maxy) = minmax $ map coords path
		         coords :: CPoint -> Point
			 coords (_,x,y) = (x,y)
			 code = concatMap (drawCP . shift) path
	                 size = round2 $ add1 (g (maxx,maxy)) 20
			 shift (col,x,y) = (g (x,y),col)
			 drawCP (p,col) = rectangle col p (2,2)
			 g (x,y) = ((x-minx)*hor+40,(y-miny)*ver+20)
			 			 			    
zeta,eta,expc,two,quadC,phylloc :: ComplexFun
zeta z@(x:+y)  = sum $ 1:[(float n:+0)**(-f x) | n <- [2..50]]
		 where f x = if x > 0 then z else (-x):+y
eta z@(x:+y)   = sum $ 1:[f (even n) $ (float n:+0)**(-g x) | n <- [2..50]]
                 where f b z = if b then -z else z
		       g x = if x > 0 then z else (-x):+y
expc z@(x:+y)  = z*exp(-f x) where f x = if x > 0 then z else (-x):+y
quadC (x:+y)   = x*x:+y*y
two (x:+y)     = (x+x):+(y+y)
sico (x:+y)    = sin x:+cos y
phylloc (x:+y) = fst (f x):+snd (f y)
		 where f x = successor p0 (sqrt x) $ x*137.50776405003785
	   
svgGraph :: Svg [CPath]
svgGraph (hor,ver) paths = (code,size)
                     where (x1,y1,x2,y2) = minmax $ concatMap coords paths
		           coords :: CPath -> Path
			   coords (_,_,ps) = ps
		           code = concatMap (drawP . shift) paths
	                   size = round2 $ add1 (f (x2,y2)) 20
			   shift (col,mode,ps) = (col,mode,map f ps) 
			   f (x,y) = ((x-x1)*hor+40,(y-y1)*ver+20)

drawP :: CPath -> String
drawP (col,mode,path@(p:ps)) = edgesCode++nodesCode
    where [nodeM,edgeM,smooth,hueM,gradM] = showMode mode
          c = hue (read [hueM]) col lg
	  is = indices path
	  nodes f code = concat $ zipWith (code . f . c) ks path
	  nodesCode = case nodeM of 
	  	      '1' -> ""
		      '2' -> nodes mkLight $ \c p -> ellipse c p (12,12)
		      '3' -> nodes mkLight $ \c p -> rectangle c p (12,12)
		      '4' -> drawBoxes c path $ map show is
		      '5' -> nodes id $ \c p -> ellipse c p (6,6)
		      _ -> nodes id $ \c p -> rectangle c p (6,6)
	  edges b = fst $ fold2 g ("",p) ps ks where
	            g (code,p) q i = (code++code',q)
	                          where code' | b    = polygon smooth (c i) tria
	                                      | True = line (c i) p q
	                                tria = [p,q,gravity path]
          edgesCode = case edgeM of '1' -> ""
                                    '2' -> edges False
				    '3' -> edges True
	  			    '4' -> polyline (smooth > '1') col path
				    _ -> polygon smooth col path
          (ks,lg) = case gradM of '1' -> (is,lg1)
			          '2' -> sym
			          '3' -> h angle
				  _   -> h slope
                    where lg1 = length ps-1
		          lg2 = lg1`div`2
		          half = [0..lg2-1]
		          sym = if lg1`mod`2 == 0 
		                then (half++reverse half,lg2)
			        else (half++lg2:reverse half,lg2+1)
		          h rel = (map f rels,length set)
	                   where rels = fst $ foldl g ([],p) ps
				 f rel = case search (== rel) set of Just i -> i
			         				     _ -> 0
                                 g (is,p) q = (is++[rel p q],q)
			         set = sort (<=) $ mkSet rels

svgLGraph :: Int -> Svg [CLPath]
svgLGraph mode (hor,ver) paths = (code,size)
          where (x1,y1,x2,y2) = minmax $ concatMap coords paths
                coords :: CLPath -> Path
                coords (_,path) = map (\(_,x,y) -> (x,y)) path
                code = concatMap (drawLP . shift) paths
                size = round2 $ add1 (f (x2,y2)) 20
                shift (col,ps) = (col,[(lab,x',y') | (lab,x,y) <- ps, 
			                             let (x',y') = f (x,y)])
	        f (x,y) = ((x-x1)*hor+40,(y-y1)*ver+20)	
	        g (lab,p,q) = (lab,(p,q))
	        drawLP :: CLPath -> String
	        drawLP (col,path) = polyline False col ps ++ drawBoxes c ps labs
	                            where (labs,ps) = unzip $ map g path
		                          c = hue mode col $ length ps

-- CURVES

data Curve = C {file :: String, colors :: [RGB], modes :: [Int],
		paths :: [Path]} deriving Show
		   
emptyC file = C file [] [] []

single :: String -> RGB -> Int -> Path -> Curve
single file col mode pa = C file [col] [mode] [pa]

drawC ::  Point -> Curve -> IO ()
drawC sc (C file cols modes paths) 
			      = do writeFile file $ show $ zip3 cols modes paths
				   readAndDraw sc file "" svgGraph $ Just . read

drawCS :: Bool -> String -> [Curve] -> IO ()
drawCS frame dir curves = do mkHtml frame dir n
		             drawDir p1 dir "" (const svgGraph) (map f curves) n
                          where n = length curves
	                        f (C _ cols modes paths) = zip3 cols modes paths

drawGraph :: Point -> String -> Int -> IO ()
drawGraph sc file mode = readAndDraw sc file "" (svgLGraph mode) $ Just . read
		
mapCurve :: (Point -> Point) -> Curve -> Curve
mapCurve f c = c {paths = map (map f) $ paths c}

updFile :: String -> Curve -> Curve
updFile file c = c {file = file}

updCol :: RGB -> Curve -> Curve
updCol col c  = c {colors = take (length $ colors c) $ repeat col}

hueCol :: Int -> Curve -> Curve
hueCol mode c = c {colors = zipWith f (colors c) $ indices ps}
	        where ps = paths c
		      f col = hue mode col $ length ps
			    
shift :: Point -> Curve -> Curve
shift (a,b) c = c {paths = map (map $ \(x,y) -> (x+a,y+b)) $ paths c}

scale,hscale,vscale,turn :: Float -> Curve -> Curve
scale a c  = c {paths = map (map $ apply2 (a*)) $ paths c} 
hscale a c = c {paths = map (map $ \(x,y) -> (a*x,y)) $ paths c} 
vscale a c = c {paths = map (map $ \(x,y) -> (x,a*y)) $ paths c} 
turn a c   = turn0 a (gravity $ concat $ paths c) c

turn0 :: Float -> Point -> Curve -> Curve
turn0 a p c = c {paths = map (map $ \q -> rotate p a q) $ paths c} 

updMod,shiftCol,takeC,dropC :: Int -> Curve -> Curve
updMod mode c = c {modes = take (length $ modes c) $ repeat mode}
shiftCol n c  = c {colors = map ((!!n) . iterate nextCol) $ colors c}
takeC n c     = c {paths = map (take n) $ paths c}
dropC n c     = c {paths = map (drop n) $ paths c}

complCol,flipH,flipV,transpose,toCenter :: Curve -> Curve
complCol   = shiftCol 765
flipH      = mapCurve $ \(x,y) -> (x,-y)
flipV      = mapCurve $ \(x,y) -> (-x,y)
transpose  = mapCurve $ \(x,y) -> (y,x)
toCenter c = shift (apply2 negate $ gravity $ concat $ paths c) c
toStart c  = shift (apply2 negate $ head $ head $ paths c) c

combine :: [Curve] -> Curve
combine [c]      = c
combine cs@(c:_) = c {colors = concatMap colors cs, modes = concatMap modes cs,
		      paths = concatMap paths cs} 
combine _        = emptyC "combine"

overlay = combine . map toCenter

inCenter :: (Curve -> Curve) -> Curve -> Curve
inCenter f c = shift (gravity $ concat $ paths c) $ f $ toCenter c

morphing :: Int -> Int -> [Curve] -> Curve
morphing mode n cs = combine $ concat $ zipWith f (init cs) $ tail cs
 where f c d = map g [0..n]
	       where g i = c {colors = map hc $ colors c,
	                      paths = zipWith h (paths c) $ paths d}
	                   where h ps qs = zipWith morph ps' qs'
		                           where diff = length qs-length ps
					         ps' = adaptLength ps diff
					         qs' = adaptLength qs $ -diff
			         morph (xc,yc) (xd,yd) = (next xc xd,next yc yd)
			         next x z = (1-inc)*x+inc*z
			         inc = float i/float n
				 hc col = hue mode col n i 
	
adaptLength :: Path -> Int -> Path
adaptLength ps n = if n > 0 then addPoints ps $ dps/2:replicate (n-1) dps
                            else ps
                   where dps = perimeter ps/k; k = float n

rainbow = transform hue
shine   = transform $ \mode col n i -> lights col n $ mode*i

transform :: (Int -> RGB -> Int -> Int -> RGB) -> Int -> Int -> Curve -> Curve
transform tr mode n c = combine $ map f [0..n-1]
			where f i = c {colors = map g $ colors c,
		                       paths = map h $ paths c}
				    where g col = tr mode col n i
				          h ps = map (\(x,y) -> (a*x,a*y)) ps
	                                         where a = float (n-i)/float n

rect :: RGB -> Int -> Point -> Curve
rect col mode p@(b,h) = single "rect" col mode [p,(b,-h),(-b,-h),(-b,h),p]
		
arc :: RGB -> Int -> Int -> [Float] -> Int -> Curve
arc col mode n rs k = single "arc" col mode $ last ps:ps
	              where ps = fst $ foldl f ([],0) $ take k $ cycle rs
			    f (ps,a) 0 = (ps,a+inc)
			    f (ps,a) r = (successor p0 r a:ps,a+inc)
			    inc = 360/float (n*length rs)
			    
poly :: RGB -> Int -> Int -> [Float] -> Curve
poly col mode n rs = arc col mode n rs $ n*length rs 

tria,circ :: RGB -> Int -> Float -> Curve
tria col mode r = turn 30 $ poly col mode 3 [r]
circ col mode r = poly col mode 100 [r]

elli :: RGB -> Int -> Float -> Float -> Curve
elli col mode a b = vscale (b/a) $ circ col mode a

piles :: RGB -> Int -> [Int] -> Curve
piles col mode s = single "piles" col mode $ coords s 0
	          where maxheight = 24*(length s-1)
	                coords (n:s) x = mkPile n x maxheight ++ coords s (x+24)
	                coords _ _     = []
	           	mkPile 0 _ _ = []
	           	mkPile n x y = float2 (x,y):mkPile (n-1) x (y-24)

cant,snake,phyllo :: RGB -> Int -> Int -> Curve
cant col mode n = single "cant" col mode $ map (float2 . p) [0..n*n-1]
	          where p 0 = (0,0)
	    	        p i = if even x 
		              then if even y
			           then if x > 0 
			                then if y' < n then (x-1,y') else (x',y)
					else if y' < n then (0,y') else (1,y) 
			           else if x' < n then (x',y-1) else (x,y') 
		              else if even y
			           then if y > 0 
			                then if x' < n then (x',y-1) else (x,y') 
					else if x' < n then (x',0) else (x,1)
			           else if y' < n then (x-1,y') else (x',y) 
		              where (x,y) = p $ i-1; x' = x+1; y' = y+1
			  
snake col mode n  = single "snake" col mode $ map (float2 . p) [0..n*n-1]
	            where p i = (if even yi then xi else n-1-xi,yi)
			        where xi = i`mod`n; yi = i`div`n
			    
phyllo col mode n = single "phyllo" col mode $ map f [1..n]
		    where f i = successor p0 (sqrt k) $ k*137.50776405003785
		    				       -- golden angle
			        where k = float i

leaf :: RGB -> RGB -> Int -> Float -> Int -> Curve
leaf col col' mode r k = combine [toStart $ c col,toStart $ flipH $ c col']
                         where a = 0.18*(501-float k)
		               c col = turn a $ arc col mode 1000 [r] k
		
blosLeaf :: RGB -> Int -> Bool -> Int -> [Int] -> Curve
blosLeaf col mode double n ks = turtle col 0 $ map (Put . toStart) cs 
   where lg = n*length ks
         (cs,_,_) = foldl f ([],col,0) $ take lg $ cycle ks
	 f (cs,col,a) 0 = (cs,col,a+inc)
         f (cs,col,a) k = (turn a (leaf col col' mode 100 k):cs,next col',a+inc)
			  where col' = if double then next col else col
         next = if double then nextColor $ 2*lgc else nextColor lgc
	 lgc = n*length (filter (/= 0) ks)
	 inc = 360/float lg
		
blosCurve :: RGB -> Int -> (RGB -> Curve) -> Curve
blosCurve col n c = turtle col 0 $ map (Put . toStart) cs 
                  where (cs,_,_) = iterate f ([],col,0)!!n
		        f (cs,col,a) = (turn a (c col):cs,nextColor n col,a+inc)
		        inc = 360/float n

-- knight paths (Springer-Wege)

kpath :: RGB -> Int -> Int -> Float -> Float -> Curve
kpath col mode n a b = case pathloop n (a,b) of 
                            Just ps -> single "kpath" col mode ps
			    _ -> emptyC "kpath"
			
pathloop :: Int -> Point -> Maybe Path
pathloop n p = f (n*n) [p] [nextPoss n [] p]
	   where f 1 ps _            = Just $ reverse ps
	         f k ps ((p:qs):pss) = if p `elem` ps then f k ps qss
			               else f (k-1) (p:ps) $ nextPoss n ps p:qss
			               where qss = qs:pss
                 f k ps pss          = do _:ps <- Just ps; _:pss <- Just pss
			   		  f (k+1) ps pss

nextPoss,newSucs :: Int -> Path -> Point -> Path
nextPoss n = sortVal . newSucs n
newSucs n visited = filter (`notElem` visited) . sucs n

sucs :: Int -> Point -> Path
sucs n p = [q | q@(a,b) <- incrs p, 0 < a, a <= float n, 0 < b, b <= float n]
		    
incrs :: Point -> Path
incrs (x,y) = [(x+1,y+2),(x+1,y-2),(x-1,y+2),(x-1,y-2),
	       (x+2,y+1),(x+2,y-1),(x-2,y+1),(x-2,y-1)]
	      --[(x+1,y),(x,y-1),(x-1,y),(x-1,y),
	      -- (x+1,y+1),(x+1,y-1),(x-1,y+1),(x-1,y-1)]
		    
sortVal :: (Point -> Path) -> Point -> Path
sortVal f = sort r . f where r x y = length (f x) < length (f y)

-- knight tours (Springer-Kreise)

ktour :: RGB -> Int -> Int -> Float -> Float -> Curve
ktour col mode n a b = case tourloop n (n*n) [p] [nextPoss n [] p] of 
		     	    Just ps -> single "ktour" col mode ps
			    _ -> emptyC "ktour"
		       where p = (a,b)

tourloop :: Int -> Int -> Path -> [Path] -> Maybe Path
tourloop n 1 ps@(p:qs) (_:pss) = if q `elem` sucs n p then Just $ reverse $ q:ps 
						      else tourloop n 2 qs pss
	          	         where q = last ps            -- = init
tourloop n k ps ((p:qs):pss)   = if p `elem` ps then tourloop n k ps qss
           		         else tourloop n (k-1) (p:ps) $
	                                       nextPosst n ps (last ps) p:qss
			         where qss = qs:pss           -- last ps = init
tourloop n k ps pss            = do _:ps <- Just ps; _:pss <- Just pss
				    tourloop n (k+1) ps pss

nextPosst :: Int -> Path -> Point -> Point -> Path
nextPosst n visited init p
       | init `notElem` sucs n p = case singlesuc fp of
	                           Just i -> case singlesuc $ context fp i of
			                     Just _ -> []   		-- (c)
					     _ -> [fp!!i]		-- (b)
                                   _ -> sortVal f p			-- (d)
       | null $ f init 	         = []                    		-- (a)
       | True 		         = sortVal f p			        -- (d)
    			           where f = newSucs n visited
			                 fp = f p
				         singlesuc = search ((==1) . length . f)
       -- (a)-(d) implement the synonymous rules of 
       -- Rabhi, Lapalme, Algorithms, Addison-Wesley 1999, Exercise 8.7.

-- SNOWFLAKES

-- snow huemode mode d n k c computes a Koch snowflake with turn mode in 
-- {1,..,6}, depth n and the k copies of scale(i/3)(c) at level 
-- 1<=i<=n revolving around each copy of scale((i-1)/3)(c) at level i-1. 
-- The real number d should be taken from the closed interval [-1,1]. 
-- d affects the radia of the circles consisting of k copies of w.

snow :: Int -> Int -> Float -> Int -> Int -> Curve -> Curve
snow huemode mode d n k c = 
     if n <= 0 || null (paths c) then emptyC "snow"
     				 else color $ combine $ c:cs
     where color = if huemode < 4 then id else hueCol $ huemode-3
           ps = concat $ paths c
           a = 360/float k
	   cs = g (n-1) [p0] $ maximum $ map (distance $ gravity ps) ps
	   g :: Int -> Path -> Float -> [Curve]
	   g 0 _ _  = []
	   g i ps r = zipWith3 h qs angs flips++g (i-1) qs r'
             where qs = concatMap circle ps
	           r' = r/3
		   circle p@(x,y) = if odd mode then s else p:s
		            where s = take k $ iterate (rotate p a) (x,y-r+d*r')
		   f col = hue huemode col n $ n-i
		   h p a b = shift p $ turn a $ scale (b/3^(n-i)) c'
		   c' = if huemode < 4 then c {colors = map f $ colors c} else c
           angs | mode < 5  = zeros 
	        | mode == 5 = iter++angs
		| True      = 0:iter++concatMap f [1..n-2]
		              where f i = concatMap (g i) iter 
		      	            g 1 a = a:iter
				    g i a = g k a++f k where k = i-1
	   iter  = take k $ iterate (+a) 0
	   zeros = 0:zeros
	   ones  = 1:ones
	   blink = 1: -1:blink
	   flips = case mode of 1 -> blink
	                        2 -> 1:take k blink++flips
				_ -> ones

-- CURVES FROM ACTIONS

data Action = Turn Float | Move Float | Jump Float | Put Curve | 
	      Open RGB Float Bool | Close | Draw | Skip deriving Show
		
-- instance Eq Action where Put (C _ _ _ ps) == Put (C _ _ _ qs) = ps == qs
--			    _ == _ = False

executeActs :: [Action] -> Path
executeActs = fst . foldl f ([(0,0)],0) where
              f (ps,a) (Move d) = (ps++[successor (last ps) d a],a)
              f (ps,a) (Turn b) = (ps,a+b)
              f psa _           = psa

heActs = Turn (-120):Move 90:concat(replicate(5)[Turn 60,Move 90])
hsActs = Turn (-135):acts++Turn 45:acts
         where acts = [Move 90,Turn 90,Move 90,Turn 45,Move 90]
pyActs = [Turn (-90),act,Turn 38.659805,Move 57.628117,Turn 91.14577,
	  Move 70.292244,Turn 50.19443,act,Turn 90,act] 
	 where act = Move 90

exec1 = executeActs heActs
exec2 = executeActs hsActs
exec3 = executeActs pyActs
              
mkActs :: Path -> [Action]
mkActs = f 0 where f a (p:ps@(q:_)) = Turn (b-a):Move (distance p q):f b ps
	  		     	      where b = angle p q
                   f _ _ = []

turtle :: RGB -> Int -> [Action] -> Curve
turtle col mode acts = case foldl f iniState acts of 
			    (c,(_,_,ps,col,b):_) -> g c ps col b
			    _ -> emptyC "turtle"
      where iniState = (emptyC "turtle",[(0,1,[p0],col,smooth)])
	    str = showMode mode; smooth = str!!2 == '2'
	    mode' b = if b && not smooth then read $ updList str 2 '2' else mode
	    f st@(c,states@((a,sc,ps,col,b):s)) act =
	      case act of Jump d          -> (g c ps col b,(a,sc,[q],col,b):s)
				             where q = successor p (d*sc) a
                          Move d          -> (c,(a,sc,ps++[q],col,b):s)
			                     where q = successor p (d*sc) a
			  Turn a'         -> (c,(a+a',sc,ps,col,b):s)
			  Skip            -> st
			  Open col sc' b' -> (c,(a,sc*sc',[p],col,b'):states)
			  Close           -> (g c ps col b,s)
			  Draw            -> (g c ps col b,(a,sc,[p],col,b):s)
			  Put c'          -> (combine [c,c1],states) where
			                  c1 = turn0 a p $ shift p $ scale sc c'
	      where p = last ps
	    g c ps col b = if length ps < 2 then c 
		 	   else combine [c,single "turtle" col (mode' b) ps]

-- FRACTALS

data Direction = North | East | South | West

go   = Move 90 
up   = Turn (-90)
down = Turn 90

north,east,south,west :: [Action]
north = [up,Move 5,down]
east  = [Move 5]
south = [down,Move 5,up]
west  = [Move $ -5]

hilbertActs :: Int -> Direction -> [Action]
hilbertActs 0 = const []
hilbertActs n = \case East -> hSouth++east++hEast++south++hEast++west++hNorth
                      West -> hNorth++west++hWest++north++hWest++east++hSouth
                      North -> hWest++north++hNorth++west++hNorth++south++hEast
                      South -> hEast++south++hSouth++east++hSouth++north++hWest
                where h = hilbertActs (n-1); hEast = h East; hWest = h West
                      hNorth = h North; hSouth = h South
		 
hilb :: RGB -> Int -> Int -> Curve
hilb col mode = turtle col mode . flip hilbertActs East

fibo :: RGB -> Int -> Int -> Float -> Float -> Curve
fibo col mode n a b  = turtle col mode $ concatMap f $ take n fib
                       where fib  = concat fibs
		             fibs = [0]:zipWith (++) fibs ([1]:fibs)
			     f 0 = [Turn a,Move 5]
			     f _ = [Turn (-b),Move 5]
			 
bush,dragon,dragonz,fern,fernL,fernU,gras,grasL :: RGB -> Int -> Int -> Curve

bush col mode n = turtle col mode $ up:f n
  	          where f 0 = [Move 3]
                        f i = acts++acts++open:Turn 20:acts++acts++Close:
                              open:Turn (-20):acts++Turn 20:acts++[Close]
			      where acts = f (i-1)
	                open = Open col 1 True

bushL col mode n = turtle col mode $ up:f n col
  	           where f 0 _   = [Move 3]
                         f i col = acts++acts++open:Turn 20:acts++acts++Close:
                                   open:Turn (-20):acts++Turn 20:acts++[Close]
			           where acts = f (i-1) $ nextColor n col
	                                 open = Open col 1 True

dragon col mode n = turtle col mode $ f n
  	            where f 0 = [move]
                          f i = f (i-1)++down:move:g (i-1)
	                  --f i = turn:f (i-1)++Turn (-135):g (i-1)++[turn]
			  --g 0 = [turn,move]
			  --g i = f (i-1)++turn:g (i-1)++[Turn (-45)]
  			  g 0 = [move]
			  g i = f (i-1)++up:move:g (i-1)
			  move = Move 5
			  turn = Turn 45

dragonz col mode n = flipH $ turtle col mode $ take (n+n) drag
	             where drag  = zip blink drag
		           zip (x:m:s) t = x:m:zip t s
			   blink = up:move:down:move:blink
			   move = Move 5
			
fern col mode n = turtle col mode $ up:f n 1
            where f 0 _ = [Move 10]
                  f i a = g 0.35 (a*50) (-a)++g 0.35 (-a*50) a++g 0.85 (a*2.5) a
	                  where g sc a b = Move 10:Open col sc False:Turn a:
		                           f (i-1) b++[Close]
 
fernL col mode n = turtle col mode $ up:f n 1 col
        where f 0 _ _   = [Move 10]
              f i a col = g 0.35 (a*50) (-a)++g 0.35 (-a*50) a++g 0.85 (a*2.5) a
	                  where g sc a b = Move 10:Open col sc False:Turn a:
		                           f (i-1) b (nextColor n col)++[Close]
 
fernU col mode n = turtle col mode $ up:f n 0
  	           where f 0 _ = []
                         f n 0 = act:open:Turn 30:g 1++Close:
		                 open:Turn (-30):g 1++Close:act:g 0
	                         where act = Move $ 1.5^(n-1)
		                       g = f (n-1)
                         f n k = f (n-1) $ k-1
			 open = Open col 1 False
 
fernUL col mode n = turtle col mode $ up:f n 0 col
  	            where f 0 _ _   = []
                          f n 0 col = act:open:Turn 30:g 1++Close:
		                      open:Turn (-30):g 1++Close:act:g 0
	                              where act = Move $ 1.5^(n-1)
		                            g k = f (n-1) k $ nextColor n col
					    open = Open col 1 False
                          f n k col = f (n-1) (k-1) $ nextColor n col

gras col mode n = turtle col mode $ up:f n
  	          where f 0 = [Move 6]
                        f i = m:open False:h (-25)++m:open False:h 37.5++
	                      open True:m:h 12.5 
			      where m = Move $ 2^i 
			            h a = Turn a:f (i-1)++[Close]
		        open = Open col 1

grasL col mode n = turtle col mode $ up:f n col
  	           where f 0 _   = [Move 6]
                         f i col = m:open False:h (-25)++m:open False:h 37.5++
	                           open True:m:h 12.5 
			   where m = Move $ 2^i 
			         h a = Turn a:f (i-1) (nextColor n col)++[Close]
				 open = Open col 1

grasC :: RGB -> Int -> Int -> Curve -> Curve
grasC col mode n c = turtle col mode $ open 6:up:f n++[Close]
   	            where f 0 = [Put $ scale 0.2 c]
                          f i = m:up:open 1:acts++Close:open 1:down:acts++Close:
		                down:m:open 1:down:m:acts++Close:up:acts
	                        where acts = f $ i-1;    m = Move $ 2^i
                                      up = Turn $ -22.5; down = Turn 22.5	
	                  open sc = Open col sc False

bunch :: Int -> Curve
bunch n = turtle red 12111 $ if n <= 0 then []
	  			       else up:f (-60)++f (-30)++f 0++f 30++f 60
          where f a = [Open red 1 False,Turn a,go,Put $ poly red 13111 6 [6,12],
	               down,Open red 0.75 False,Put $ bunch $ n-1,Close,Close]

-- based curves

triaA,square,quad,penta,hexa,pytrunk,cactus,rhombLB,rhombRB,rhombLS,rhombRS 
	                                                  :: RGB -> Int -> Curve
	       
triaA c   = flip (single "triaA" c)  [p0,(45,-77.94229),(90,0),p0]
square c  = flip (single "square" c) [p0,(0,-90),(90,-90),(90,0),p0]
quad c    = flip (single "quad" c) [p0,(22,-55),(80,-55),(70,22),p0]
penta c   = flip (single "penta" c) [p0,(-27.811533,-85.595085),(45,-138.49576),
				     (117.81153,-85.595085),(90,0),p0]
hexa c    = flip (single "hexa" c)  [p0,(-45,-77.94228),(0,-155.88458),
 				     (90,-155.88458),(135,-77.94229),(90,0),p0]				    
pytrunk c = flip (single "pytrunk" c) [p0,(0,-90),(36,-135),(90,-90),(90,0),p0]
cactus c  = flip (single "cactus" c)  [p0,(7.5,-60),(22.5,-90),(67.5,-90),
				       (82.5,-60),(90,0),p0]
rhombLB c = flip (single "rhombLB" c) [p0,(-27.811533,-85.595085),
				       (62.18847,-85.595085),(90,0),p0]
rhombRB c = flip (single "rhombRB" c) [p0,(27.811533,-85.595085),
				       (117.81152,-85.595085),(90,0),p0]
rhombLS c = flip (single "rhombLS" c) [p0,(-72.81154,-52.900665),
				       (17.188461,-52.900665),(90,0),p0]
rhombRS c = flip (single "rhombRS" c) [p0,(72.81154,-52.900673),
				       (162.81152,-52.900673),(90,0),p0]

base :: Curve -> Curve
base c = if null pss || length ps < 3 || p /= last ps || d == 0 then c 
         else c {paths = map mkBased pss}
	 where pss@(ps@(p:qs):qss) = paths c
	       q = last $ init qs; d = distance p q; a = -angle p q
	       mkBased = map $ apply2 (*(90/d)) . rotate p0 a . sub2 p
				    
octo :: RGB -> Int -> Curve
octo col mode = poly col mode 8 [5]

-- grow c bunch docks the based curves given by bunch at c.
				
grow :: Curve -> [[Action]] -> [Action]
grow c bunch = Put c:(concat $ zipWith g bunch $ mkLines $ concat $ paths c)
          where g [] _             = []
                g acts (p@(x,y),q) = Open red 1 False:Jump x:down:Jump y:Turn a:
	   			     Open red (d/90) False:acts++[Close,Close]
				     where a = angle p q-90; d = distance p q

-- CURVE TREES

data CurveTree = CT Curve [CurveTree]

one c = CT c [] 

mt = one $ emptyC "curve"

buildCT :: Curve -> [Bool] -> Int -> CurveTree
buildCT c bs = f where f 0 = one c
		       f n = CT c $ map g bs where g True = f $ n-1 
		       				   g _    = mt

depth (CT _ [])  = 1
depth (CT _ cts) = maximum (map depth cts)+1

evalCT :: CurveTree -> Curve
evalCT t@(CT c _) = turtle red 1 $ f (depth t) t
                    where f :: Int -> CurveTree -> [Action]
                          f 0 _          = []
                          f i (CT c cts) = grow c $ map (f $ i-1) cts 

evalCTC :: (Curve -> Curve) -> CurveTree -> Curve
evalCTC trans t@(CT c _) = turtle red 1 $ f n (colors c!!0) t
                  where n = depth t
                        f :: Int -> RGB -> CurveTree -> [Action]
                        f 0 _ _            = []
                        f i col (CT c cts) = grow (trans $ updCol col c) $ 
                                             map (f (i-1) $ nextColor n col) cts 

-- repeat based curves 

trias,squares,hexas,pytree,pytreeC :: RGB -> Int -> (Curve -> Curve) -> Int 
								     -> Curve
trias c mode trans   = evalCTC trans . buildCT (triaA c mode) (replicate 3 True)
squares c mode trans = evalCTC trans . buildCT (square c mode) 
					       (replicate 5 True)
hexas c mode trans   = evalCTC trans . buildCT (hexa c mode) (replicate 6 True)
pytree c mode trans  = evalCT . buildCT (trans $ pytrunk c mode) 
			                [False,True,True]
pytreeC c mode trans = evalCTC trans . buildCT (pytrunk c mode) 
					       [False,True,True]

octos :: Curve
octos = scale 0.4 $ evalCT $ buildCT (base $ octo black 14111) 
				     (replicate 8 True) 4
					   
trunk1 = evalCT $ CT (pytrunk red 15111) [mt,one $ flipH $ pytrunk blue 15111,
					     one $ pytrunk blue 15111]

-- GRAPHIC PRIMITIVES

angle :: RealFloat a => (a,a) -> (a,a) -> a
angle (x1,y1) (x2,y2) = f (y2-y1) (x2-x1)*180/pi where f 0 0 = atan2 0 1
			      			       f x y = atan2 x y

slope :: Point -> Point -> Float
slope (x1,y1) (x2,y2) = if x1 == x2 then float maxBound else (y2-y1)/(x2-x1) 
			      
-- successor p (distance p q) (angle p q) = q. 

successor :: Point -> Float -> Float -> Point
successor (x,y) r a = (x+r*c,y+r*s) where (s,c) = sincos a
				 -- successor p 0 _ = p
				 -- successor (x,y) r 0 = (x+r,y) 
				 -- successor (x,y) r a = rotate (x,y) a (x+r,y)

sincos a = (sin rad,cos rad) where rad = a*pi/180 	-- sincos 0 = (0,1)

distance :: Point -> Point -> Float
distance (x1,y1) (x2,y2) = sqrt $ (x2-x1)^2+(y2-y1)^2

perimeter :: Path -> Float
perimeter ps = if peri <= 0 then 0.01 else peri
               where peri = sum $ zipWith distance ps $ tail ps
	        
mkLines :: Path -> [(Point,Point)]
mkLines ps = zip ps $ tail ps 
		
addPoints :: Path -> [Float] -> Path
addPoints (p:ps@(q:_)) (d:ds) = if d > d' then p:addPoints ps (d-d':ds)
 		                else p:addPoints (successor p d a:ps) ds
		                where d' = distance p q; a = angle p q
addPoints ps _ = ps

-- rotate q a p rotates p clockwise by a around q on the axis (0,0,1).
-- rotate (0,0) 45 (5,5) --> (0.0,7.071068) = (0.0,sqrt $ 5^2+5^2)

rotate :: Point -> Float -> Point -> Point
rotate _ 0 p             = p 			
rotate q@(i,j) a p@(x,y) = if p == q then p else (i+x1*c-y1*s,j+x1*s+y1*c)
			   where (s,c) = sincos a; x1 = x-i; y1 = y-j

	
-- rotateA q (a,nx,ny,nz) p rotates p clockwise by a around q on the axis
-- (nx,ny,nz).

-- rotateA p0 45 (0,0,1) (5,5) --> (0.0,7.071068)
-- rotateA p0 45 (1,0,0) (5,5) --> (5.0,3.535534)
-- rotateA p0 45 (0,1,0) (5,5) --> (3.535534,5.0)

rotateA :: Point -> Float -> (Float,Float,Float) -> Point -> Point
rotateA _ 0 _ p 		     = p
rotateA q@(i,j) a (nx,ny,nz) p@(x,y) = if p == q then p 
				       else (f i (c'*nx*nx+c) (c'*nx*ny-s*nz),
      		                             f j (c'*nx*ny+s*nz) (c'*ny*ny+c))
                                       where (s,c) = sincos a; c' = 1-c
                                             f i a b = i+(x-i)*a+(y-j)*b

spline :: Path -> Path
spline ps@(p:_:_:_) = if p == last ps then spline0 True $ init ps
				      else spline0 False ps
spline ps = ps
					  
-- spline0 b ps uses ps as control points for constructing a closed (b = True) 
-- resp. open (b = False) B-spline with degree 3; see Paul Burke, Spline Curves
-- (http://astronomy.swin.edu.au/~pbourke/curves/spline)
-- or Heinrich Mller, B-Spline-Technik, Vorlesung Geometrisches Modellieren 
-- (http://ls7-www.cs.tu-dortmund.de).

spline0 :: Bool -> Path -> Path
spline0 b ps = first:map f [1..resolution] ++ map g [1..9] ++
	       [if b then first else ps!!(n-1)]
         where first = f 0; n = length ps; resolution = n*6
	       f i = point $ upb*float i/float resolution 
               g i = point $ upb+float i/10
	       upb = float n-if b then 1 else 3
	       point v = foldl1 add2 $ map h [0..n-1] 
	                 where h i = apply2 (*z) $ ps!!i
	                         where z | b && v < u i = blend2 u i $ v-u 0+u n
		                         | b 	        = blend2 u i v
		     	                 | True         = blend2 t i v 
               t i = if i < 3 then 0 else float (min i n)-2
	       u i = if i <= n then float i else u (i-1)+u (i-n)-u (i-n-1)
	       h d s = if d == 0 then 0 else s
	       blend2 t i v = h denom1 sum1+h denom2 sum2
	       		      where ti = t i; ti3 = t $ i+3
		                    denom1 = t (i+2)-ti;  num1 = v-ti
				    denom2 = ti3-t (i+1); num2 = ti3-v
				    sum1 = num1/denom1*blend1 t i v
				    sum2 = num2/denom2*blend1 t (i+1) v
               blend1 t i v = h denom1 sum1+h denom2 sum2
		              where ti = t i; ti1 = t $ i+1; ti2 = t $ i+2
		                    denom1 = ti1-ti;  num1 = v-ti 
				    denom2 = ti2-ti1; num2 = ti2-v
				    sum1 = if b i then num1/denom1 else 0
				    sum2 = if b $ i+1 then num2/denom2 else 0
				    b i = t i <= v && v < t (i+1)

-- COLORS

data RGB = RGB Int Int Int deriving Eq

red     = RGB 255 0 0
green   = RGB 0 255 0
blue    = RGB 0 0 255
magenta = RGB 255 0 255
cyan    = RGB 0 255 255
yellow  = RGB 255 255 0 
orange  = RGB 255 180 0
brown   = RGB 0 160 255
black   = RGB 0 0 0
white   = RGB 255 255 255
grey    = RGB 100 100 100

lights :: RGB -> Int -> Int -> RGB
lights (RGB x y z) n i = RGB (f x) (f y) (f z)
             where f x | i > 0 = x+i*(255-x)`div`n -- i=n  --> f x = 255 (white)
		       | True  = x+i*x`div`n       -- i=-n --> f x = 0   (black)

light col = lights col 2 1 

lighter col = lights col 3 2 
			  
mkLight :: RGB -> RGB
mkLight c = if c == black then white else light c

-- nextCol computes the successor of each color c on a chromatic circle of 1530
-- equidistant pure (or hue) colors. A color c is pure if c is neither black nor
-- white and at most one of the R-, G- and B-values of c is different from 0 and
-- 255.

nextCol,complColor :: RGB -> RGB
 
nextCol = \case RGB 255 n 0 | n > 0   -> RGB 255 (n-1) 0  -- n = 255 --> yellow
		RGB 255 0 n | n < 255 -> RGB 255 0 (n+1)  -- n = 0   --> red
		RGB n 0 255 | n > 0   -> RGB (n-1) 0 255  -- n = 255 --> magenta
		RGB 0 n 255 | n < 255 -> RGB 0 (n+1) 255  -- n = 0   --> blue 
		RGB 0 255 n | n > 0   -> RGB 0 255 (n-1)  -- n = 255 --> cyan 
		RGB n 255 0 | n < 255 -> RGB (n+1) 255 0  -- n = 0   --> green
		c  		      -> c 

(^^^) :: (a -> a) -> Int -> a -> a                             
f^^^n = if n == 0 then id else f . (f^^^(n-1))
        -- = (!!n) . iterate f = foldl1 (.) $ replicate n f

complColor = nextCol^^^765 
 
-- hue mode col n i computes the i-th successor of col in a chromatic circle of 
-- n <= 1530 pure colors.

hue :: Int -> RGB -> Int -> Int -> RGB
hue 1 col n i = iterate nextCol col!!round (float i*1530/float n) 
hue 2 col n 0 = col
hue 2 col n i = if odd i then complColor $ hue 2 col n $ i-1
                         else nextColor (n `div` 2) $ hue 2 col n $ i-2
hue 3 col n i = if odd i then complColor d else d where d = hue 1 col n i
hue 4 col n i = if odd i then complColor d else d where d = hue 1 col (2*n) i
hue _ col _ _ = col

nextColor :: Int -> RGB -> RGB
nextColor n col = hue 1 col n 1

instance Read RGB where 
         readsPrec _ s = [(RGB r g b,s) | ("RGB",s) <- lex s, (r,s) <- reads s,
    		                          (g,s) <- reads s, (b,s) <- reads s] ++
                         [(rgb c,s)     | (c,s) <- lex s, c `elem` 
                         words "red green blue yellow magenta cyan black white"]
		  
rgb = \case "red" -> red; "green" -> green; "blue" -> blue; "yellow" -> yellow
	    "magenta" -> magenta; "cyan" -> cyan; "black" -> black
	    "white" -> white

instance Show RGB where 
              showsPrec _ (RGB x y z) = ("RGB "++) . shows x . (' ':) . shows y
						             . (' ':) . shows z

-- BASIC SVG CODE GENERATORS

rectangle :: RGB -> Point -> Point -> String
rectangle col (x,y) (rx,ry) = polygon '1' col [(x1,y1),(x2,y1),(x2,y2),(x1,y2)]
                              where x1 = x-rx; x2 = x+rx; y1 = y-ry; y2 = y+ry
		    
drawBox :: (RGB -> Point -> Point -> String) -> RGB -> Point -> String -> String
drawBox f col p@(x,y) a = if null a then "" 
			  else f col p (width+2,12) ++ 
			       text black (x-width,y+5) (concatMap g a)
                          where width = case a of "&#955;" -> 5
                                        	  "&#128520;" -> 10
                                        	  _ -> lgW a
                                g '\\' = "&#92;"
                                g c    = [c]

drawBoxes :: (Int -> RGB) -> Path -> [String] -> String
drawBoxes c ps = concat . zipWith3 (drawBox ellipse . mkLight . c) [0..] ps

ellipse :: RGB -> Point -> Point -> String
ellipse (RGB r g b) (x,y) (rx,ry) = "\n<ellipse cx"&show x  ++ " cy"&show y ++ 
				              " rx"&show rx ++ " ry"&show ry ++ 
					      " fill"&("rgb"++show (r,g,b)) ++
				    "/>"

line :: RGB -> Point -> Point -> String
line (RGB r g b) (x1,y1) (x2,y2) = "\n<line x1"&show x1 ++ " y1"&show y1 ++ 
				          " x2"&show x2 ++ " y2"&show y2 ++ 
					  " stroke"&("rgb"++show (r,g,b)) ++
					  " stroke-width"&".6" ++ "/>"

polyline :: Bool -> RGB -> Path -> String
polyline smooth (RGB r g b) ps = if null ps then ""
			         else "\n<polyline points"&
				      concatMap (f . round2) qs ++
				      " stroke"&("rgb"++show (r,g,b)) ++
				      {- " stroke-width"&".6" ++ -}
				      " fill"&"none" ++ "/>"
		                 where f (x,y) = ' ':show x++',':show y
				       qs = if smooth then spline ps else ps

polygon :: Char -> RGB -> Path -> String
polygon smooth (RGB r g b) ps = if null ps then ""
		                else "\n<polygon points"&
				     concatMap (f . round2) qs ++ 
				     " fill"&("rgb"++show (r,g,b)) ++ "/>"
		                where f (x,y) = ' ':show x++',':show y
			              qs = case smooth of 
			                        '1' -> ps
			                        '2' -> spline $ last ps:ps 
			                        _ -> spline ps

text :: RGB -> Point -> String -> String
text (RGB r g b) (x,y) a = "\n<text x"&show x ++ " y"&show y ++ 
			   " font-family"&"courier" ++ " font-size"&"16" ++ 
			   " fill"&("rgb"++show (r,g,b)) ++ '>':f a ++"</text>" 
			   where f "<" = "&lt;"; f ">" = "&gt;"; f "&" = "&amp;"
			         f str = str
			{- " fill"&("rgb"++show (r,g,b)) ++ '>':concatMap f a ++ 
			   "</text>" where f '<' = "&lt;"; f '>' = "&gt;"
			   		-- f '&' = "&amp;"
			   		   f c = [c] -}

-- SAMPLE PICTURES

curve1 = drawC p1 $ rect red 15511 (111,222)
curve2 = drawC p1 $ scale 5 $ morph43 111
curve3 = drawC p1 $ snow1 6 15111
curve4 = drawC p1 $ rainbow 1 3 $ snow1 6 15111
curve5 = drawC p1 $ morphing 1 3 $ map ($ snow1 6 15111) [id,scale 0]

drawPiles :: [Int] -> IO ()					   
drawPiles = drawC p0 . piles red 31121	            		     

drawCastles :: [[Int]] -> IO ()					   
drawCastles = drawCS True "piles" . map (piles red 31131)	            		     

schedule = drawGraph p0 "fahrplan" 				-- hor=4, ver=20
								   
orbits = combine $ concat $ take 5 $ iterate (map $ scale 0.5)     -- hor=ver=33
	           [f 8 4,turn 45 $ f 8 4,f 4 8,turn 45 $ f 4 8]       
         where f = elli blue 12111
         
orbitsM = morphing 1 10 [orbits,scale 0.3 orbits]

hilb1  = hilb red 12111 6
hilb2  = hilb red 21111 4					   -- hor=ver=24
hilb3  = hilb red 31111 4		
hilb4  = hilb red 31121 4		
hilb5  = hilb red 14211 4		
hilb6  = hilb red 13111 4		
hilb7  = hilb red 42111 4

hilb8  = hilb red 22121 4					   -- hor=ver=33
hilb9  = hilb red 12112 4		
hilb10 = hilb red 12113 4		
hilb11 = hilb red 12114 4		
hilb12 = flipV $ toCenter hilb1
hilb13 = turn 22 $ hilb1
hilb14 = overlay [hilb1,hilb2]

cant1 = cant red 12111 33
cant2 = flipV $ toCenter cant1
cant3 = overlay [cant1,cant2]
cant4 = cant red 12121 33
cant5 = cant red 13111 33
cant6 = cant red 13113 33
cant7 = cant red 13114 33

snake1 = snake red 12111 33
snake2 = transpose snake1
snake3 = overlay [snake1,snake2]
snake4 = snake red 12211 33

cs1 = overlay [cant3,snake1]					   -- hor=ver=15
cs2 = overlay [cant1,snake3]
cs3 = overlay [cant3,snake3] 					
cs4 = updMod 12121 cs3 		

-- see http://www.win.tue.nl/~hzantema/strfib.html

fibo1 = flipH $ toCenter $ fibo10 21111 			   -- hor=ver=88
fibo2 = flipH $ toCenter $ fibo10 31111
fibo3 = flipH $ toCenter $ fibo10 12111
fibo4 = flipH $ toCenter $ fibo10 14111
fibo5 = flipH $ toCenter $ fibo10 14211
fibo6 = flipH $ toCenter $ fibo10 15111
fibo7 = flipH $ toCenter $ fibo10 13124
fibo8 = flipH $ toCenter $ turn 72 $ fibo red 13124 500 144 72  
fibo9 = turn 225 $ fibo yellow 12111 4150 18 162
fibo10 m = turn 72 $ dropC 100 $ fibo red m 500 144 72
fibo11 m = fibo red m 2000 30 150
fibo12 m = fibo red m 2000 90 120
fibo13 m = fibo red m 2000 20 160
fibo14 m = fibo red m 5000 174 174

cf1 = overlay [cant3,scale 5 $ fibo3]				   -- hor=ver=18
cf2 = overlay [cant5,flipV $ toCenter cant5,scale 5 fibo7]
cf3 = overlay [cant5,flipV $ toCenter cant6,scale 5 fibo7]
cf4 = overlay [cant6,flipV $ toCenter cant6,scale 5 fibo7] 	

phyllo1 = phyllo red 21111 300					   -- hor=ver=14
phyllo2 = phyllo red 41111 300	
phyllo3 = phyllo yellow 21111 600
phyllo4 = phyllo yellow 21113 600
phyllo5 = phyllo yellow 21114 600
phyllo6 = phyllo yellow 21124 600
phyllo7 = phyllo yellow 21134 600
		     
rhombH col = single "rhombH" col 15211 [p0,(3,-2),(16,0),(3,2),p0]

blosR col = blosCurve col 6 rhombH

gras1 = hueCol 1 $ gras red 14111 8				   -- hor=ver=1
gras2 = grasL red 14111 7
gras3 = hueCol 1 $ updCol red $ grasC red 14111 5 $ rhombH green   
gras4 = hueCol 1 $ updCol red $ grasC red 14111 5 $ blosR green
gras5 = hueCol 1 $ grasC red 14111 5 $ blosR green
gras6 = hueCol 1 $ grasC red 14111 5 $ hueCol 1 $ blosR green
gras7 = grasC green 14111 5 $ blosR red

fern1 = hueCol 1 $ fern red 14111 8				   -- hor=ver=4
fern2 = fernL red 14111 8
fern3 = hueCol 1 $ fernL red 14111 8	

fern4 = hueCol 1 $ fernU red 14111 11				   -- hor=ver=3
fern5 = fernUL red 14111 11					   
fern6 = hueCol 1 $ fernUL red 14111 11
fern7 = fernUL red 12111 11 
fern8 = hueCol 1 $ fernU red 12111 12
fern9 = hueCol 1 $ fernU red 12111 14

bush1 = hueCol 1 $ bush red 14111 5				   -- hor=ver=2
bush2 = bushL red 14111 5				           

dragon1 = dragon green 12111 11					   -- hor=ver=1
dragon2 = dragonz green 12111 4225				   -- hor=ver=2
dragon3 = dragonz green 13213 3333

sp1 = combine [snow1 4 12111,scale 0.7 phyllo1] 		   -- hor=ver=12
sp2 = shiftCol 500 sp1			
sp3 = shiftCol 1000 sp1			

circ1 = circ red 13111 33
circ2 = circ red 15111 33
rect1 = rect red 15211 (45,5) 
rect2 = rect red 13211 (5,45)

rain1     = rainbow 1 33 circ1
rain2     = rainbow 1 33 circ2 
rain2'    = morphing 1 33 [circ2,scale 0 circ2]
rain3     = rainbow 1 33 rect1
rain4     = rainbow 1 33 rect2
rain5 m n = rainbow m n $ tria red 13111 33
rain6 m n = rainbow m n $ tria red 13113 33

rads1 = [1..9]++reverse [2..8]
rads2 = [2,2,3,3,4,4,5,5,4,4,3,3]

arc1 = arc red 12111 10 [33,44] 5				   -- hor=ver=4
arc2 = arc red 12111 10 [33,44] 11 				
arc3 = arc red 12111 10 [33,44] 19 				

poly1 = poly red 12111 11 rads1					   -- hor=ver=33
poly2 = poly red 13111 11 rads1
poly3 = poly red 14111 11 rads1
poly4 = poly red 14211 11 rads1
poly5 = poly red 13121 11 rads2
poly6 = combine [poly2,turn (-1.5) poly5] 

poly7 = poly red 15211 30 [40]
poly8 = poly red 15211 5 [5,40]
poly11 = morphing 1 44 [poly7,poly8]
poly12 = turn 36 $ scale 0.5 poly8
poly13 = poly red 12111 2 [0.1]
poly14 m = turn 54 $ poly cyan m 12 [210,70]

leaf1 = leaf red green 15111 111 1000
leaf2 = leaf red green 15111 111 500
leaf3 = leaf red green 15111 111 200

blos1 = blosLeaf red 15111 True 6 [200,150]			   -- hor=ver=3
blos2 = blosLeaf red 15111 True 6 [100,0,0,80,0,60,0,80,0,0]
blos3 = blosLeaf red 15111 True 6 [200,0]
blos4 = blosLeaf red 15111 False 6 [200,0]

cp1 = combine [cant3,scale 0.15 poly1]
cp2 = combine [updMod 12121 $ cant3,scale 0.15 poly1]
cp3 = combine [updMod 12121 $ cant3,scale 0.25 poly6] 		   -- hor=ver=18
cp4 = overlay [cant4,flipV $ toCenter cant4,scale 0.25 poly6]
cp5 = combine [f blue,flipH $ f yellow]
      where f col = hueCol 1 $ updCol col $ bunch 4

hexas1 = hexas red 15111 id 1
hexas2 = hexas red 15111 (inCenter $ rainbow 1 33) 1
	   			      
hexas6 mode f = scale 0.3 $ evalCT t1 -- mode in {13111,13211,14211,15111,15311}
		where c col = f $ hexa col mode
		      t1  = CT (c red) $ replicate 6 t2
		      t2  = CT (c green) [mt,mt,t31,t32]
		      t31 = CT (c blue) [mt,mt,t41,t42]
		      t32 = CT (c blue) [mt,mt,t43]
		      t41 = CT (c magenta) [mt,mt,t51,t52]
		      t42 = one $ c magenta
		      t43 = CT (c magenta) [mt,t51,t53]
		      t51 = CT (c yellow) [mt,mt,t6,t6]
		      t52 = one $ c yellow
		      t53 = CT (c yellow) [mt,mt,t6]
		      t6  = one $ c cyan
		      
hexas6r  = hexas6 15111 $ inCenter $ rainbow 1 33
hexas6re = hexas6 15111 $ rainbow 1 33
hexas6s  = hexas6 15111 $ inCenter $ shine 1 33
hexas6d  = hexas6 15111 $ inCenter $ shine (-1) 33
hexas6se = hexas6 15111 $ shine 1 33
hexas6h  = hueCol 1 $ hexas6 15111 id
                          
hexasR n col mode f = scale 0.3 $ evalCTC f t1
		      where hex  = CT (hexa col mode)		
                            t1   = hex $ replicate 6 t2		
                            t2   = hex [mt,mt,t $ n-3,u $ n-3]
                            t 0  = hex [mt,mt,leaf,leaf]	
                            t n  = hex [mt,mt,v $ n-1,leaf]	
                            u 0  = hex [mt,mt,leaf]			
                            u n  = hex [mt,mt,w $ n-1]			
                            v 0  = hex [mt,mt,leaf,leaf]	
                            v n  = hex [mt,mt,t $ n-1,leaf]		
                            w 0  = hex [mt,leaf,leaf]			
                            w n  = hex [mt,t $ n-1,u $ n-1]			
                            leaf = hex []		

hexas12 = hexasR 12 magenta 15111 $ inCenter $ rainbow 1 33

sizeH1 n = length $ paths $ hexasR n red 15111 id 	-- sizeH1 = sizeH2
sizeH2 n = 1+sum (map (6*) [1..n])			-- sizeH1 60 = 10981

hexasRl n col = hexasR n col 12111 id
hexasRt n col = hexasR n col 13111 id
hexasRp n col = hexasR n col 15111 id
hexasRr n col = hexasR n col 15111 $ inCenter $ rainbow 1 33
hexasRs n col = hexasR n col 15111 $ inCenter $ shine 1 33
hexasRh mode n col = hueCol mode $ hexasR n col 15111 id
hexasRu mode n col = hueCol mode $ hexasR n col 13111 id
hexasRg mode n col = hueCol mode $ hexasR n col 13211 id

act0 = drawC p1 $ morphing 1 30 [c,scale 0.2 $ turn 60 c]
       where c = hexasRu 1 6 red

grow0  = pytree red 15111 (inCenter $ rainbow 1 33) 6
grow0C = pytreeC red 15111 (inCenter $ rainbow 1 33) 6

grow1 mode = scale 0.4 $ evalCT t1
             where t1 = CT (pytrunk green mode) 
             		   [mt,t2,mt,CT (hexa blue mode)[mt,t3]]
	           t2 = one $ pytreeC cyan mode id 4
	           t3 = one $ pytreeC magenta mode id 6
		   
grow2 mode = evalCT t1
             where t1 = CT (inCenter (shine 1 33) $ base $ quad yellow mode) 
                           [t2,t3,t4,t5]
                   t2 = one $ inCenter (shine (-1) 33) $ penta green mode
		   t3 = one $ rain $ flipH $ base $ poly red mode 7 [33]
		   t4 = one $ pytreeC yellow mode rain 3
		   t5 = one $ rain $ hexa blue mode
                   rain = inCenter $ rainbow 1 33
	   
grow3 = grow7 id id id 
	   
grow4 = grow7 (rainbow 1 5) id id 
	   
grow5 = grow7 (shine 1 5) (shine 1 11) id
	   
grow6 = grow7 (shine 1 5) (shine 1 11) $ rainbow 1 5

grow7 f g h mode = inCenter f $ overlay [evalCT t1,evalCT t2]
            where t1 = CT (flipH $ base $ poly yellow mode 16 [111])
	        		        $ concat $ replicate 8 [mt,t3]
                  t2 = CT (base $ poly yellow mode 16 [111])
	        		$ concat $ replicate 8 [mt,t3]
                  t3 = CT (inCenter g $ cactus blue mode) $ replicate 5 t4
		  t4 = one $ shift (45,0) $ inCenter h $ elli green mode 22 44
					    
grow8 mode1 mode2 = evalCT t1
                    where c1 = poly red mode1 5 [222]
                          c2 = poly red mode1 5 [111]
                          c3 = flipH $ base $ snow 1 mode2 1 4 5 $ turn 54 c1
                          t1 = CT c3 [one $ hexa blue mode1,t2,
                                      one $ flipH $ base c2,one c3,one c3]
                          t2 = CT (square red mode1) [mt,one c3]

grow9  = evalCT $ CT tr [mt,CT tr [mt,one tr,one tr],one tr]
	 where tr = rainbow 1 51 $ pytrunk red 15111

grow10 = evalCTC (rainbow 1 51) $ CT tr [mt,CT tr [mt,one tr,one tr],one tr]
	 where tr = pytrunk red 15111

morph1 n    = morphing 1 n [rect red 15111 (222,222),circ red 15111 22]
morph2 n    = turn 18 $ morphing 1 n [poly8,poly12,turn 72 poly8]
morph3 n    = updMod 23121 $ morph2 n
morph4 n    = morphing 1 n [poly7,poly13]
morph5 n    = morphing 1 n [turn 45 $ poly red 12111 4 [55],tria red 12111 33]
morph6 n    = morphing 1 n [cant3,scale 0.15 poly1]
morph8 n    = morphing 1 n [rainbow 1 33 rect1,rainbow 1 33 rect2]
morph9 m n  = morphing m n $ map (rainbow 1 33 . rect yellow 15211)
			         [(5,45),(45,5)]
morph10 m n = morphing m n $ map (rainbow 1 33 . rect cyan 12114) 	-- !!
			         [(50,30),(0,30),(50,0)]
morph11 m n = morphing m n $ map (rainbow 1 33 . rect cyan 15211) 
			         [(50,25),(1,22),(50,5)]
morph12 m n = morphing m n $ map ($ poly red 13111 12 [2,12]) [id,shift (50,0)]
morph13 m n = rainbow m n circ1 	   -- morphing m n [circ1,scale 0 circ1]
morph14 m n = rainbow m n circ2 	   -- morphing m n [circ2,scale 0 circ2]
morph15 m n = morphing m n [circ1,shift (15,15) $ scale 0 circ1]	
morph16 m n = morphing m n [circ2,shift (15,15) $ scale 0 circ2]
morph17 m n = morphing m n $ map ($ snow1 6 15111) [id,shift (10,0) . scale 0]
morph18 m n = rainbow m n $ snow1 6 15111                
morph19 n   = morphing 1 n $ map ($ snow1 6 15111) [hscale 0.5,vscale 0.5]
morph20 n   = morphing 1 n $ map ($ snow1 6 15111) [hscale 0.5,id,vscale 0.5]
morph21 n   = morphing 1 n [poly red 13111 50 [44],poly red 13111 5 [1,44]]
morph22 n   = morphing 1 n [rect red 13111 (33,33),tria red 13111 40]
morph23 n   = morphing 1 n [turn 30 $ f 3,turn 45 $ f 4,turn 18 $ f 5,f 6] 
	      where f k = poly red 13111 k [40]
morph24 n   = morphing 1 n [turn 45 $ poly red 13111 4 [55],tria red 13111 33]
morph25 n   = morphing 1 n [snow33 1 6 15111,snow33 0.2 6 15111]
morph26 n   = morphing 1 n [snow34 1 6 15111,snow34 0.2 6 15111]

polys m n x y = [poly red m n [x,y],poly red m n [y,x]]

morph27 n   = morphing 1 n $ polys 15111 5 55 0.1
morph28 n   = morphing 1 n $ map ($ snow1 6 15111) [id,hscale 0.2,id,vscale 0.2]
morph30 n   = morphing 1 n $ polys 13121 5 55 0.1
morph31 n   = morphing 1 n $ polys 14121 5 55 0.1
morph32 n   = morphing 1 n $ polys 13123 5 55 0.1
morph33 n   = morphing 1 n $ polys 13124 5 55 0.1
morph34 n   = morphing 1 n $ polys 13133 5 55 0.1
morph35 n   = morphing 1 n $ polys 13114 5 55 0.1
morph36 n   = morphing 2 n [poly yellow 15111 5 [55,0.5],
 			    poly red 15111 5 [0.5,55]]
morph37 n   = morphing 3 n [poly yellow 15111 5 [55,0.5],
			    poly red 15111 5 [0.5,55]]
morph38 n   = morphing 1 n [snake1,turn 90 snake1]
morph39 n   = morphing 1 n [cant1,turn 90 cant1]
morph40 n   = morphing 1 n $ polys 15111 11 55 0.1
morph41 n   = morphing 2 n $ polys 15111 11 55 0.1
morph42 n   = morphing 1 n $ polys 15211 11 60 0.1 ++ polys 15211 11 40 0.1 ++
			     polys 15211 11 20 0.1
morph43 n   = morphing 1 n $ polys 15111 11 60 0.1 ++ polys 15111 11 40 0.1 ++
			     polys 15111 11 20 0.1

tria1 n = tria red n 222	
tria2 n = tria red n $ -222
rect3 n = rect red n (222,222)

snow1 m    = snow 1 m 1 4 6 . tria1				
snow2 m    = snow 1 m 1 4 6 . tria2
snow3 m    = shiftCol 111 . snow1 m
snow4 m n  = snow 1 m 1 4 6 $ circ cyan n 222
snow8 m    = snow 1 m 1 5 6 . updCol yellow . tria1
snow9 m n  = snow 1 m 1 4 6 $ poly red n 6 [210,70]
snow10 m n = snow 1 m 1 4 4 $ poly red n 4 [210,70]
snow11 m n = snow 1 m 1 4 4 $ poly red n 8 [210,70]
snow12 m n = snow 1 m 1 4 8 $ poly red n 4 [210,70]
snow13 m   = snow 1 m 1 4 5 . poly14
snow14 m   = snow 1 m 1 4 10 . poly14
snow15 m   = snow 1 m 1 4 15 . poly14
snow16 m   = snow 1 m 1 4 4 . rect3 
snow17 m   = snow 1 m 1 3 8 . rect3 
snow18 m   = snow 1 m 1 5 6 . rect3 				-- !!
snow19 m   = snow 1 m 1 6 6 . rect3 
snow20 m   = snow 1 m 1 5 5 . poly14 
snow21 m   = snow 1 m 1 5 4 . rect3 
snow22 m   = snow 1 m 1 4 6 . tria2 
snow23 m n = snow 1 m 1 5 6 $ combine [tria1 n,tria2 n]  	-- !!
snow24 m n = snow 1 m 1 4 6 $ circ red n 222
snow26 k m = hueCol k . updCol yellow . snow 1 m 1 4 10 . poly14
snow27 m n = snow 1 m 1 4 6 $ poly red n 4 [210]
snow28 m n = snow 1 m 1 4 6 $ poly red n 4 [210,70]
snow29 m n = snow 1 m 1 5 6 $ poly red n 4 [210,70]
snow30 m n = snow 1 m 1 4 6 $ poly red n 4 [210,105]
snow33 r m n = snow 1 m r 4 6 $ poly red n 6 [210,110]
snow34 r m n = snow 1 m r 4 5 $ turn 18 $ poly red n 5 [210,110]
snow35 m n = snow 1 m 1 5 5 $ turn 54 $ poly red n 5 [333]
snow36 m n = snow 1 m 1 4 5 $ toCenter $ pytree red n id 3
snow37 m n = snow 1 m 1 4 4 $ toCenter $ pytree red n id 3
snow38 m   = snow 1 m 1 3 5 . toCenter . grow3
snow39 m   = snow 1 m 1 5 6 . tria1				-- !!
snow40 m   = snow 1 m 1 5 6 . rect3 				-- !!
snow41     = snow 1 5 (-0.8) 5 8 . rect3
snow42     = snow 1 1 (-1)   5 8 . rect3
snow43     = snow 1 2 0.17   5 6 . rect3
snow44 n   = snow 4 2 1 5 6 $ combine $ map (tria blue n) [222,-222]
snow45 n   = snow 4 2 1 5 6 $ combine $ map (tria yellow n) [222,-222]

blos7 m = blosLeaf red m True 6 [200,155]

snow46 = snow 1 6 1 2 6       . blos7
snow47 = snow 1 6 (-0.15) 2 6 . blos7
snow48 = snow 1 6 (-0.15) 3 6 . blos7

snow49 = snow 1 4 (-1) 4 6 $ circ red 13113 222
snow50 = snow 1 4 0 4 6 $ circ red 13113 222
snow51 = snow 1 4 1 4 6 $ circ red 13113 222			-- !!
snow52 = snow 4 4 1 4 6 $ circ red 13113 222
snow53 = snow 1 6 1 5 6 $ tria1 15111
snow54 = snow 4 6 1 5 6 $ tria1 15111
snow55 = snow 1 6 1 4 6 $ rect red 15111 (222,222)		-- !!
snow56 = snow 1 6 1 4 6 $ combine $ map (tria red 13113) [222,-222]

snows k = drawCS False ("snows"++show k) $
          case k of 2 -> rains snow1
	  	    3 -> rains snow27
	            4 -> rains snow30
	            5 -> rains snow28
	            6 -> rains $ snow33 1 -- $ 1/3
	            7 -> rains $ snow34 1 -- $ 1/3
	            8 -> rains snow23
	            9 -> rains snow22
	            10 -> concatMap (rains . snow33) [0.8,0.6,0.4,0.2]
		    _ -> []
	      
rains c  = map morph [3..11] where morph n = rainbow 1 n $ c 6 15111

morphs = moviePix False "morphs" 
from3  = movies "from3" $ [(i,j) | i <- words "morph27 morph27A", j <- [0..22]]
			  ++  [("urtiere",j) | j <- [0..9]]
snows1 = movies "snows1" [(i,j) | i <- map (("snows"++) . show) [2..9], 
			          j <- cycle9 0]
snows2 = movieSelect "snows2" $ cycle9 0
snows3 = movieSelect "snows10" $ concatMap cycle9 [0,9,18,27,18,9]

cycle9 n = [n..n+8]++[n+7,n+6..n+1]

path1 = kpath red 12111 6 1 1 
path2 = kpath red 14211 8 1 1 
path3 = kpath red 14211 8 8 8 
path4 = overlay [path1,flipH $ toCenter path1,path2,flipH $ toCenter path2]
path5 = kpath red 12111 20 1 1 
path6 = kpath red 14211 40 1 1 
path7 = kpath red 12113 40 40 40 

{- kpath 6 1 1 -->
   Just [(1,1), (2,3), (1,5), (3,6), (5,5), (6,3), (5,1), (3,2), (1,3), (2,1),
         (4,2), (6,1), (5,3), (6,5), (4,6), (3,4), (2,6), (1,4), (2,2), (4,1),
	 (6,2), (5,4), (6,6), (4,5), (3,3), (2,5), (4,4), (5,6), (6,4), (5,2),
	 (3,1), (1,2), (2,4), (1,6), (3,5), (4,3)] -}

tour1 = ktour red 12211 8 1 1 					-- hor=ver=44
tour2 = ktour red 14211 8 8 8 
tour3 = overlay [tour1,flipH $ toCenter tour1,tour2,flipH $ toCenter tour2]
tour4 = ktour red 12111 16 8 8 
tour5 = ktour red 12111 18 9 9 					-- hor=ver=22
tour6 = ktour red 12113 18 9 9 
tour7 = ktour red 12114 18 9 9 
tour8 = toCenter $ ktour red 13111 18 9 9 
tour9 = hueCol 1 $ overlay [p,flipH $ toCenter p,q,flipH $ toCenter q]
        where p = ktour red 14211 8 1 1
	      q = ktour red 14211 8 8 8 

{- ktour 8 1 1 -->
   Just [(1,1), (2,3), (3,1), (1,2), (2,4), (1,6), (2,8), (4,7), (6,8), (8,7),
         (7,5), (8,3), (7,1), (5,2), (7,3), (8,1), (6,2), (4,1), (2,2), (1,4), 
	 (2,6), (1,8), (3,7), (5,8), (7,7), (8,5), (6,6), (7,8), (8,6), (7,4), 
	 (8,2), (6,1), (4,2), (2,1), (3,3), (5,4), (3,5), (4,3), (5,1), (6,3),
	 (8,4), (7,2), (6,4), (5,6), (4,8), (2,7), (1,5), (3,6), (1,7), (3,8), 
	 (5,7), (4,5), (5,3), (6,5), (4,6), (6,7), (8,8), (7,6), (5,5), (3,4), 
	 (1,3), (2,5), (4,4), (3,2), (1,1)] -}
