module Tableau where

-- 8.7.2022

import Control.Monad
import Data.Maybe
import Painter(updList,Parser,Tree(F,V),isV,root,subtrees,nodes,leaves,term,
	       parse,token,tstring,tchar,many,char,digit,nat,newlines,fold2,
	       foldTree,drawTermsP,mkTrue,mkFalse)

-- getSub t p is the subtree of t at position p by u. 
-- putSub t u replaces it by u.

getSub :: Tree a -> [Int] -> Tree a
getSub t [] = t
getSub (F _ ts) (i:is) | i < length ts = getSub (ts!!i) is
getSub _ _  = error "getSub"

putSub :: Tree a -> [Int] -> Tree a -> Tree a
putSub _ [] u          = u
putSub (V a) [0] u     = F a [u]
putSub (F a ts) (n:p) u 
   | n < lg            = F a $ updList ts n $ putSub (ts!!n) p u
   | n == lg && null p = F a $ ts++[u] where lg = length ts
putSub _ _ _           = error "putSub"

-- path(t)(p) is the list of labels on the path from the root of t to p.

path :: Tree a -> [Int] -> [a]
path t p = map (root . getSub t . flip take p) [0..length p]

-- GRAMMAR
{-
Formula    ::= OrExpr | OrExpr "==>" OrExpr
OrExpr     ::= AndExpr ("\\/" AndExpr)*
AndExpr    ::= NotExpr ("/\\" NotExpr)*
PrefixExpr ::= ("not" | "EX" | "AX") PrefixExpr | BasicExpr
BasicExpr  ::= "(" OrExpr ")" | VarExpr | ConstantExpr
VarExpr    ::= "x"Nat | "y"Nat | "z"Nat
ConstExpr  ::= "false" | "true"
-}

-- PARSER FOR FORMULAS

formula,basic :: Parser (Tree String)
formula = do t <- aux; msum [do tstring "==>"; u <- aux; return $ F "==>" [t,u],
		             return t]
          where aux = prop "\\/"
basic = msum [token $ do x <- msum $ map char "xyz"; n <- many digit
			 return $ V $ x:n, 
	      do c  <- msum $ map tstring $ words "true false"; return $ F c [],
	      do op <- msum $ map tstring $ words "not EX AX"; a <- basic
	         return $ F op [a],
              do tchar '('; t <- formula; tchar ')'; return t] 

prop :: String -> Parser (Tree String)
prop op = do t <- aux op; ts <- many $ do tstring op; aux op
 	     return $ if null ts then t else F op $ t:ts
 	  where aux "\\/" = prop "/\\"
 	        aux _     = basic

-- TABLEAU CALCULUS

type Tableau = Tree TabNode
data TabNode = TN {state :: Int, phi :: Tree String, visited :: Bool} 
	       deriving Eq

isBox (F "AX" _)           = True
isBox (F "not" [F "EX" _]) = True
isBox _                    = False

not_ t = F "not" [t]

contra :: TabNode
contra = TN 0 mkFalse False

iniTab :: Tree String -> Tableau
iniTab phi = V $ TN (if (foldTree f) phi then 0 else 1) phi False
	     where f "EX" _ = False
		   f "AX" _ = False
		   f _ bs   = and bs

-- TABLEAU REDUCER

reduceTab :: String -> String -> IO ()
reduceTab file str = case parse formula str of
                          Just phi -> do writeFile redfile $ newlines $ map f $
					           iteration (iniTab phi) 2 []
                                         drawTermsP redfile
                          _ -> putStrLn "reduceTab undefined"
                     where redfile = file ++ "Reduction" 
                           f (t,next,p) = show (t,p)   
			       
iteration :: Tableau -> Int -> [Int] -> [(Tableau,Int,[Int])]
iteration t n p = case redStepP t n of Just (u,k,q) | (t,n) /= (u,k) 
		       	                 -> prev:(t,n,q):iteration u k q
		                       _ -> [prev]
		  where prev = (t,n,p)

-- In addition to redStep (see Painter.hs), redStepP stores the root positions 
-- of the redices and reducts of all reduction steps.

redStepP :: Tableau -> Int -> Maybe (Tableau,Int,[Int])
redStepP t n = f t n [] where
               f u n p = msum [do (t,n) <- tabRules u n t p
		 		  Just (t,n,p),
		 	       do F a ts <- Just u
		 	          (ts,n,p) <- g ts n p 0
		 	          Just (F a ts,n,p)]
	       g ts n p i = do t:ts <- Just ts
		               case f t n $ p++[i] of 
				    Just (t,n,p) -> Just (t:ts,n,p)
		 		    _ -> do (ts,n,p) <- g ts n p $ i+1
					    Just (t:ts,n,p)

tabRules :: Tableau -> Int -> Tableau -> [Int] -> Maybe (Tableau,Int)
tabRules tab succ t p =
 do guard $ not visited
    msum [do F "\\/" phis <- Just phi
	     add succ [(p++[n], tnF phi) | p <- ps,(n,phi) <- zip [0..] phis],
          do F "/\\" phis <- Just phi
             add succ [(p++q, tnF phi) | p <- ps, (q,phi) <- zip zeroPoss phis],
          do F "==>" [phi,psi] <- Just phi
	     add succ $ [(p++[0], tnF $ not_ phi) | p <- ps] ++
	     	        [(p++[1], tnF psi)        | p <- ps],
          do F "EX" [phi] <- Just phi
             add (succ+1) $ [(p++[0], tnF $ F (show succ) []) | p <- ps] ++
	     		    [(p++[0,0], TN succ phi False)    | p <- ps] ++
            	            [(p++q, TN succ phi False)        | p <- ps, 
				    (q,phi) <- zip (drop 2 zeroPoss) $ boxes p],
              
          do F "AX" [phi] <- Just phi
	     add succ [(p++q, TN succ phi False) | p <- ps, 
					    (q,succ) <- zip zeroPoss $ succs p],
          do F "not" [F "not" [phi]] <- Just phi
	     Just (tab' $ tnF phi, succ),
          do F "not" [F "EX" [phi]] <- Just phi
	     Just (tab' $ tnF $ F "AX" [F "not" [phi]], succ),
          do F "not" [F "AX" [phi]] <- Just phi
	     Just (tab' $ tnF $ F "EX" [F "not" [phi]], succ),
          do F "not" [F "\\/" phis] <- Just phi
             add succ [(p++q, tnF $ not_ phi) | p <- ps,
	     				        (q,phi) <- zip zeroPoss phis],
          do F "not" [F "/\\" phis] <- Just phi
	     add succ [(p++[n], tnF $ not_ phi) | p <- ps, 
	                                          (n,phi) <- zip [0..] phis],
          do F "not" [F "==>" [phi,psi]] <- Just phi
	     add succ $ [(p++[0], tnF phi)          | p <- ps] ++
	     	        [(p++[0,0], tnF $ not_ psi) | p <- ps],
          do z@(V x) <- Just phi
	     let ps = [p | p <- nodes tab, 
	     		   root (getSub tab p) == tnF (F "not" [z])]
             guard $ not $ null ps; Just (mkFalse ps, succ),
          do F "not" [z@(V x)] <- Just phi
             let ps = [p | p <- nodes tab, root (getSub tab p) == tnF z]
             guard $ not $ null ps; Just (mkFalse ps, succ)]
 where TN state phi visited = root tab
       tnF phi = TN state phi False
       tab' root | isV tab = V root 
                 | True    = F root $ subtrees tab
       ps = [p | p <- leaves tab, root (getSub tab p) /= contra]
       mkFalse = foldl f tab where f t p = putSub t p $ V contra
       add succ nodes = Just (fold2 putSub (tab' new) ps $ map V labs, succ)
	                where new | isBox phi = TN state phi True
	                   	  | True      = TN state mkTrue True
			      (ps,labs) = unzip nodes
       zeroPoss = iterate (0:) [0]
       -- boxes(q) and succs(q) are the sets of all state formulas 
       -- (state,[]phi,True) resp. transitions (state,state') on the path from
       -- the root of t to p++q.
       boxes :: [Int] -> [Tree String]
       boxes q = [phi | TN state' (F "AX" [phi]) True <- path t $ p++q, 
	 		state == state']
       succs :: [Int] -> [Int]
       succs q = [fromJust succ | TN state' (F x []) _ <- path t $ p++q, 
	 		          state == state',
	 		          let succ = parse nat x, isJust succ]

-- UNPARSER FOR TABLEAU NODES

instance Show TabNode where 
         show (TN _ (F "false" []) _)   = "\"false\""
         show (TN _ (F "true" []) _)    = "\"true\""
         show (TN state (F succ [] ) _) 
              | isJust $ parse nat succ = quote state $ "->" ++ 's':succ
         show (TN 0 phi b)     = '\"':addVisited phi b ++ "\""
         show (TN state phi b) = quote state $ " sats " ++ addVisited phi b

quote state str = "\"s" ++ show state ++ str ++ "\""

addVisited phi b = if b then showFormula phi ++ "%visited" else showFormula phi				     

showFormula,enclose :: Tree String -> String
showFormula (F x [t])     = x ++ ' ':enclose t
showFormula (F x (t:ts))  = enclose t ++ concatMap (f . enclose) ts
			    where f str = ' ':x++' ':str
showFormula (F x _) = x
showFormula (V x)   = x
showFormula _       = error "showFormula"
enclose t           = if isV t || root t `elem` words "not EX AX"
	              then showFormula t else '(':showFormula t++")"

-- EXAMPLES

-- EX (AX) is the modal diamond (box) operator as defined in
-- https://fldit-www.cs.tu-dortmund.de/~peter/LogikPadR.pdf.
-- The page numbers below refer to reference [38] of that report.

tabAL1 = reduceTab "tabAL1" 					     -- hor = 11
	   "(x /\\ y \\/ z) /\\ x1 \\/ y1 \\/ z1 \\/ not x2"         -- ver = 44
tabAL2 = reduceTab "tabAL2" 
	   "not x /\\ (y \\/ x) /\\ (y ==> x) \\/ x /\\ (not x /\\ y \\/ not y)" 
                   			  			     -- page 148
tabML1 = reduceTab "tabML1" "EX EX x /\\ AX (x ==> AX not x)"	     -- page 158
tabML2 = reduceTab "tabML2" "EX x /\\ AX not y /\\ EX (x ==> y)"     -- page 159
tabML3 = reduceTab "tabML3" "AX (x ==> y) /\\ EX x /\\ AX not y"     -- page 161
tabML4 = reduceTab "tabML4" "not (AX (x ==> y) /\\ EX x /\\ AX not y)"
		  					   -- negation of tabML3

